import voxelfarm
import os
import os.path
import pandas as pd

class block_model_framework:

    def __init__(self):  
        pass

    def input_string(self, id, label, default = ""):
        return ""

    def log(self, message):
        pass

    def download_entity_file(self, filename, id = None):
        return ""

    def attach_file(self, filename, id = None):
        pass

    def get_entity_folder(self, id = None):
        return ""

    def get_entity_file_list(self, id = None):
        return []

class block_model_host:

    def __init__(self, framework = None, entity = None, file = None):    
        if framework:
            self.lambda_framework = framework
        else:
            if voxelfarm.voxelfarm_framework:
                self.lambda_framework = voxelfarm.voxelfarm_framework
            else:
                self.lambda_framework = block_model_framework()

        self.df_blockmodel = None
        self.modified_attributes = []
        self.modified_gm = False

        if entity and file:
            self.blockmodel_entity = entity
            self.blockmodel_name = file
        else:
            self.blockmodel_entity = self.lambda_framework.input_string('bm_entity', 'Block Model Entity', '')
            self.lambda_framework.log('Block Model entity: ' + self.blockmodel_entity)
            self.blockmodel_name = self.lambda_framework.input_string('bm_file', 'Block Model File', '')
            self.lambda_framework.log('Block Model name: ' + self.blockmodel_name)

        gm_file = self.lambda_framework.download_entity_file(self.blockmodel_name, self.blockmodel_entity)
        self.lambda_framework.log('Block Model file: ' + gm_file)
        if os.path.isfile(gm_file):
            self.df_blockmodel = pd.read_feather(gm_file)
        else:
            self.df_blockmodel = pd.DataFrame()

    def get_geological_model(self):
        return self.lambda_framework.download_entity_file(self.blockmodel_name, self.blockmodel_entity)

    def get_file_model(self, attribute):
        if attribute in self.df_blockmodel.columns:
            return self.get_geological_model()
        else:
            return self.lambda_framework.download_entity_file(attribute + ".ftr", self.blockmodel_entity)

    def get_geological_model_name(self):
        return self.blockmodel_name

    def get_pandas(self, vars = None):
        df_model = self.df_blockmodel.copy()
        valid = True
        if vars:
            attributes = []
            for attribute in vars:
                if not attribute in attributes:
                    attributes.append(attribute)
                    if not attribute in df_model:
                        attribute_file = self.lambda_framework.download_entity_file(attribute + ".ftr", self.blockmodel_entity)
                        if os.path.isfile(attribute_file):
                            df_attribute = pd.read_feather(attribute_file)
                            df_model = pd.concat([df_model, df_attribute], axis=1)
                        else:
                            self.lambda_framework.log('Block Model attribute file ' + attribute + ' not found')
                            valid = False

            if valid:
                df_model = df_model[attributes]
            else:
                return pd.DataFrame()
            
        return df_model                

    def put_pandas(self, df_model, vars = None):
        entity_folder = self.lambda_framework.get_entity_folder(self.blockmodel_entity)
        if not os.path.exists(entity_folder):
            os.mkdir(entity_folder)

        update_gm = False
        if vars:
            for attribute in vars:
                if attribute in df_model:
                    if attribute in self.df_blockmodel:
                        update_gm = True
                    else:
                        df_attribute = df_model[[attribute]]

                        previous_attribute_file = self.get_file_model(attribute)
                        if os.path.isfile(previous_attribute_file):
                            df_previous_attribute = pd.read_feather(previous_attribute_file)
                            if len(df_attribute) != len(df_previous_attribute):
                                df_previous_attribute.update(df_attribute)
                                df_attribute = df_previous_attribute
                        elif len(df_attribute) != len(self.df_blockmodel):
                            df_model = self.df_blockmodel.copy()
                            df_model[attribute] = -999
                            df_model = df_model[[attribute]]
                            df_model.update(df_attribute)
                            df_attribute = df_model

                        attribute_file = os.path.join(entity_folder, attribute + ".ftr")
                        df_attribute.to_feather(attribute_file)
                        if attribute_file not in self.modified_attributes:
                            self.modified_attributes.append(attribute_file)
        else:
            if self.df_blockmodel.empty:
                self.df_blockmodel = df_model.copy()
                self.modified_gm = True
            else:
                update_gm = True

        if update_gm:
            self.modified_gm = True
            self.df_blockmodel.update(df_model)

    def field_list(self):
        fields = []
        bm_files = self.lambda_framework.get_entity_file_list(self.blockmodel_entity)

        for col in self.df_blockmodel.columns:
            fields.append(col)

        for file in bm_files:
            if (file != self.blockmodel_name) and file.endswith(".ftr"):
                file_name = os.path.basename(file).split('.')[0]
                fields.append(file_name)

        return fields      

    def write_tag(self, tag):
        entity_folder = self.lambda_framework.get_entity_folder(self.blockmodel_entity)
        tag_file_path = os.path.join(entity_folder, tag + '.tag')

        if tag_file_path not in self.modified_attributes:
            tag_file = open(tag_file_path, "w")
            tag_file.write('1')
            tag_file.close()
            self.modified_attributes.append(tag_file_path)

    def read_tag(self, tag):
        tag_file = self.lambda_framework.download_entity_file(tag + '.tag', self.blockmodel_entity)
        return os.path.isfile(tag_file)

    def close(self):
        if self.modified_gm:
            entity_folder = self.lambda_framework.get_entity_folder(self.blockmodel_entity)
            if not os.path.exists(entity_folder):
                os.mkdir(entity_folder)

            gm_file = os.path.join(entity_folder, self.blockmodel_name)
            self.df_blockmodel.to_feather(gm_file)
            self.lambda_framework.attach_file(gm_file, self.blockmodel_entity)

        for attribute in self.modified_attributes:
            self.lambda_framework.attach_file(attribute, self.blockmodel_entity)

        self.modified_attributes.clear()   
        self.modified_gm = False         




import voxelfarm
from datetime import datetime

class type:
    view = 1 << 1
    project = 1 << 2
    voxel_terrain = 1 << 3
    voxel_operations = 1 << 4
    block_model = 1 << 5
    point_cloud = 1 << 6
    voxel_generator = 1 << 7
    point_cloud_raw = 1 << 8
    heightmap_raw = 1 << 9
    block_model_raw = 1 << 10
    mesh_raw = 1 << 11
    mesh = 1 << 12
    ortho = 1 << 13
    program = 1 << 14
    folder = 1 << 15
    raw_density = 1 << 16
    indexed_density = 1 << 17
    material_tracking = 1 << 18
    voxel_mesh = 1 << 19
    drill_holes = 1 << 20
    drill_holes_raw = 1 << 21
    indexed_ortho = 1 << 22
    voxel_points = 1 << 23
    voxel_plane = 1 << 24
    wms_ortho = 1 << 25
    voxel_ortho = 1 << 26
    export = 1 << 27
    report = 1 << 28
    raw_geo_chem = 1 << 29
    geo_chem = 1 << 30
    value = 1
    set = 2

class attribute:
    none = 0
    volume = 1
    material = 1 << 1
    color = 1 << 2
    vector = 1 << 3
    normal = 1 << 4
    tangent = 1 << 5
    uv = 1 << 6
    color_difference = 1 << 7
    voxel_index = 1 << 8
    faces = 1 << 9
    merge = 1 << 10 
    disable_compression = 1 << 11
    fix_self_intersections = 1 << 12

class attribute_merge_mode:
    none = 0
    add = 1
    multiply = 2
    min = 3
    max = 4
    average = 5
    unit = 6
    grade = 7

class layer_merge_mode:
    none = 0
    min = 1
    max = 2
    override = 3

class texture:
    diffuse = 0
    normal = 1
    rgb0 = 2
    rgb1 = 3
    rgb2 = 4
    rgba0 = 1
    rgba1 = 1
    rgba2 = 1

class ColorLegend:
    gradient_step = 1.0
    gradient_min = 0.0
    gradient_max = 100.0
    gradient_attribute = ""
    static_color = (0, 0, 0)

class DrillHoleOptions:
    show_collars = True
    show_cylinders = True

class Material:
    color = (1.0, 1.0, 1.0, 1.0)
    gradient = None
    diffuse = False
    normal = False
    gradient = ""
    def serialize(self):
        result = "0,"
        result += str(self.color[0]) + ","
        result += str(self.color[1]) + ","
        result += str(self.color[2]) + ","
        result += str(self.color[3]) + ","
        result += ("1" if self.diffuse else "0") + ","
        result += ("1" if self.normal else "0") + ","
        result += self.gradient
        return result

class view_lambda_framework:
    def init(self):
        pass

    def get_voxel_size(self):
        return 0.5

    def input(self, id, label, default = 0.0):
        return 0.0

    def input_entity(self, id, label, type):
        return ''

    def input_attributes(self, id, label, type, entity):
        return ''

    def input_attribute(self, id, label, type, entity):
        return ''

    def input_query(self, id, label, entity):
        return ''

    def input_region(self, id, label):
        return ''

    def input_date(self, id, label, default = datetime.now):
        return datetime.fromtimestamp(0) 

    def input_color_legend(self, attributes, default_value):
        result = ColorLegend()
        return result

    def input_drillhole_options(self, entity, default):
        result = DrillHoleOptions()
        return result    

    def input_string(self, id, label, default = ""):
        return ''

    def input_bool(self, id, label, default = None):
        return 0

    def set_query(self, entity, query):
        pass

    def set_attributes(self, entity, attributes):
        pass

    def set_attribute_gradient(self, entity, attribute, min_value, max_value):
        pass

    def start_composite(self):
        pass

    def add_layer(self, layers, entity, merge_mode=0):
        pass

    def get_entity_type(self, entity):
        return 0

    def get_entity_name(self, entity):
        return ''

    def get_entity_date(self, entity):
        return datetime.fromtimestamp(0) #we lose the milliseconds

    def folder_contains(self, folder, entity):
        return 0

    def get_attributes(self, attributes):
        count = 0
        return range(0, count)

    def get_attribute_name(self, attributes, index):
        return ''

    def get_attribute_index(self, attributes, name):
        return 0

    def get_material(self, entity, voxel):
        return 0

    def load_mesh(self, entity, attributes, custom_attributes):
        return 0

    def mesh_intersection(self, componentA, componentB, attributes, custom_attributes):
        return 0

    def mesh_union(self, componentA, componentB, attributes, custom_attributes):
        return 0

    def mesh_complement(self, componentA, componentB, attributes, custom_attributes):
        return 0

    def load_voxels(self, entity, attributes, custom_attributes):
        return 0

    def voxels_intersection(self, componentA, componentB, attributes, custom_attributes):
        return 0

    def voxels_union(self, componentA, componentB, attributes, custom_attributes):
        return 0

    def voxels_complement(self, componentA, componentB, attributes, custom_attributes):
        return 0

    def load_points(self, entity, attributes, custom_attributes):
        return 0

    def load_blocktree(self, entity, attributes, custom_attributes):
        return 0

    def load_density(self, entity, attributes, custom_attributes):
        return 0

    def load_texture(self, entity, type):
        return 0

    def load_texture_stack(self):
        return ''

    def apply_texture(self, component, texture, type):
        return 0

    def apply_texture_stack(self, component, textures, type):
        pass

    def new_region_scope(self, region):
        return 0

    def set_field(self, v, value):
        pass

    def render(self, component, material):
        pass

    def log(self, message):
        print(message)

    def input_duplicate(self, entity):
        return 0

    def new_plane(self, a, b, c, d):
        return -1    

    def set_component_attribute_merge_mode(self, component, attribute, merge_mode):
        pass

    def export(self, component, name):
        pass

    def store(self, component):
        pass

    def get_entity_count(self):
        return 0

    def is_meta(self):
        return True

    def add_property(self, layer, key, value):
        pass

class view_lambda_host:

    version = (3,0,1,7)

    def __init__(self, framework = None):  
        if framework:
            self.lambda_framework = framework
        else:
            if voxelfarm.voxelfarm_framework:
                self.lambda_framework = voxelfarm.voxelfarm_framework
            else:
                self.lambda_framework = view_lambda_framework()

        self.entityCount = self.lambda_framework.get_entity_count()
        self.entities = range(0, self.entityCount)

        if self.lambda_framework.is_meta():
            self.voxelCount = 1   
            self.fieldCount = 1  
        else:
            self.voxelCount = 40*40*40
            self.fieldCount = 45*45*45

        self.voxels = range(0, self.voxelCount)
        self.field = range(0, self.fieldCount)

    def get_voxel_size(self):
        return self.lambda_framework.get_voxel_size()

    def input(self, id, label, default = 0.0):
        return self.lambda_framework.input(id, label, default)

    def input_entity(self, id, label, type):
        return self.lambda_framework.input_entity(id, label, type)

    def input_attributes(self, id, label, type, entity):
        return self.lambda_framework.input_attributes(id, label, type, entity)

    def input_attribute(self, id, label, type, entity):
        return self.lambda_framework.input_attributes(id, label, type, entity)

    def input_query(self, id, label, entity):
        return self.lambda_framework.input_query(id, label, entity)

    def input_region(self, id, label):
        return self.lambda_framework.input_region(id, label)

    def input_color_legend(self, attributes, default_value):
        result = ColorLegend()
        gradient_step = 1.0
        gradient_min = 0.0
        gradient_max = 100.0
        gradient_attribute = "fe"
        if default_value != None:
            gradient_step = default_value.gradient_step
            gradient_min = default_value.gradient_min
            gradient_max = default_value.gradient_max
            gradient_attribute = default_value.gradient_attribute

        legend_string = self.lambda_framework.input_color_legend(attributes, gradient_step, gradient_min, gradient_max, gradient_attribute)
        items = legend_string.split(',')
        result.gradient_attribute = items[0]
        result.gradient_min = float(items[1])
        result.gradient_max = float(items[2])
        result.gradient_step = float(items[3])
        return result

    def input_drillhole_options(self, entity, default):
        result = DrillHoleOptions()
        show_collars = 1
        show_cylinders = 1
        if default != None:
            show_collars = 1 if default.show_collars else 0
            show_cylinders = 1 if default.show_cylinders else 0
        self.lambda_framework.input_drillhole_options(entity, show_collars, show_cylinders)
        return result

    def input_date(self, id, label, default = datetime.now):
        timestamp = self.lambda_framework.input_date(id, label, 0)
        return datetime.fromtimestamp(timestamp/1000) # we lose the milliseconds

    def input_string(self, id, label, default = ""):
        return self.lambda_framework.input_string(id, label, default)

    def input_bool(self, id, label, default = None):
        return self.lambda_framework.input_bool(id, label, default)

    def set_query(self, entity, query):
        self.lambda_framework.set_query(entity, query)

    def set_attributes(self, entity, attributes):
        self.lambda_framework.set_attributes(entity, attributes)

    def set_attribute_gradient(self, entity, attribute, min_value, max_value):
        self.lambda_framework.set_attribute_gradient(entity, attribute, min_value, max_value)

    def start_composite(self):
        return self.lambda_framework.start_composite()

    def add_layer(self, layers, entity, merge_mode=layer_merge_mode.override):
        self.lambda_framework.add_layer(layers, entity, merge_mode)

    def get_entity_type(self, entity):
        return self.lambda_framework.get_entity_type(entity)

    def get_entity_name(self, entity):
        return self.lambda_framework.get_entity_name(entity)

    def get_entity_date(self, entity):
        timestamp = self.lambda_framework.get_entity_date(entity)
        return datetime.fromtimestamp(timestamp/1000) #we lose the milliseconds

    def get_entity_attributes(self, entity):
        attributes = self.lambda_framework.get_entity_attributes(entity)
        return attributes

    def folder_contains(self, folder, entity):
        return self.lambda_framework.folder_contains(folder, entity)

    def get_attributes(self, attributes):
        count = self.lambda_framework.get_attribute_count(attributes)
        return range(0, count)

    def get_attribute_name(self, attributes, index):
        return self.lambda_framework.get_attribute_name(attributes, index)

    def get_attribute_index(self, attributes, name):
        return self.lambda_framework.get_attribute_index(attributes, name)

    def start_sum(self, item, property):
        return self.lambda_framework.start_sum(item, property)

    def sum(self, sumId, value):
        self.lambda_framework.sum(sumId, value)

    def get_material(self, component, voxel):
        return self.lambda_framework.get_material(component, voxel)

    def get_channels(self, component):
        count = self.lambda_framework.get_channel_count(component)
        return range(0, count)

    def load_mesh(self, entity, attributes, custom_attributes):
        return self.lambda_framework.load_mesh(entity, attributes, custom_attributes)

    def mesh_intersection(self, componentA, componentB, attributes, custom_attributes):
        return self.lambda_framework.mesh_intersection(componentA, componentB, attributes, custom_attributes)

    def mesh_union(self, componentA, componentB, attributes, custom_attributes):
        return self.lambda_framework.mesh_union(componentA, componentB, attributes, custom_attributes)

    def mesh_complement(self, componentA, componentB, attributes, custom_attributes):
        return self.lambda_framework.mesh_complement(componentA, componentB, attributes, custom_attributes)

    def load_voxels(self, entity, attributes, custom_attributes):
        return self.lambda_framework.load_voxels(entity, attributes, custom_attributes)

    def voxels_intersection(self, componentA, componentB, attributes, custom_attributes):
        return self.lambda_framework.voxels_intersection(componentA, componentB, attributes, custom_attributes)

    def voxels_union(self, componentA, componentB, attributes, custom_attributes):
        return self.lambda_framework.voxels_union(componentA, componentB, attributes, custom_attributes)

    def voxels_complement(self, componentA, componentB, attributes, custom_attributes):
        return self.lambda_framework.voxels_complement(componentA, componentB, attributes, custom_attributes)

    def load_points(self, entity, attributes, custom_attributes):
        return self.lambda_framework.load_points(entity, attributes, custom_attributes)

    def load_blocktree(self, entity, attributes, custom_attributes):
        return self.lambda_framework.load_blocktree(entity, attributes, custom_attributes)

    def load_density(self, entity, attributes, custom_attributes):
        return self.lambda_framework.load_density(entity, attributes, custom_attributes)

    def load_texture(self, entity, type):
        return self.lambda_framework.load_texture(entity, type)

    def load_texture_stack(self):
        return self.lambda_framework.load_texture_stack()

    def apply_texture(self, component, texture, type):
        return self.lambda_framework.apply_texture(component, texture, type)

    def apply_texture_stack(self, component, textures, type):
        return self.lambda_framework.apply_texture_stack(component, textures, type)

    def new_region_scope(self, region):
        return range(0, self.lambda_framework.new_region_scope(region))

    def get_voxel_origin(self, v):
        return (0,0,0)

    def get_field_origin(self, v):
        return (0,0,0)

    def set_field(self, v, value):
        self.lambda_framework.set_field(v, value)

    def render(self, component, material):
        materialData = material.serialize() if material != None else ""
        self.lambda_framework.render(component, materialData)

    def log(self, message):
        self.lambda_framework.log(message)

    def new_material(self):
        return Material()

    def input_duplicate(self, entity):
        return self.lambda_framework.input_duplicate(entity)

    def new_plane(self, a, b, c, d):
        return self.lambda_framework.new_plane(a, b, c, d)

    def set_component_attribute_merge_mode(self, component, attribute, merge_mode):
        return self.lambda_framework.set_component_attribute_merge_mode(component, attribute, merge_mode)

    def export(self, component, name):
        self.lambda_framework.export(component, name)

    def store(self, component):
        self.lambda_framework.store(component)

    def add_property(self, layer, key, value):
        self.lambda_framework.add_property(layer, key, value)




import json
import base64
import time
import requests
from datetime import datetime, timedelta
import os
import random
import pandas as pd
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import as_completed
from io import StringIO
import io

class rest:

    OrgId = '2343243456678890'
    class entity_type:
        View = 'VIEW'
        Project = 'PROJECT'
        VoxelTerrain = 'VOXSURF'
        RealtimeVoxelTerrain = 'VOXOP'
        BlockModel = 'VOXBM'
        IndexedPointCloud = 'IDXPC'
        VoxelGenerator = 'VOXGEN'
        RawPointCloud = 'RAWPC'
        RawHeightmap = 'RAWHM'
        RawBlockModel = 'RAWBM'
        RawMesh = 'RAWMESH'
        IndexedMesh = 'IDXMESH'
        VoxelMesh = 'VOXMESH'
        OrthoImagery = 'ORTHO'
        IndexedOrthoImagery = 'IDXORTHO'
        Program = 'PROGRAM'
        Folder = 'FOLDER'
        RawDensity = 'RAWDENSITY'
        IndexedDensity = 'IDXDENSITY'
        MaterialTracking = 'VOXMT'
        MaterialTrackingOperation = 'VOXMTOP'
        RawDrillHoles = 'RAWDH'
        DrillHoles = 'IDXDH'
        OrthoVox='ORTHOVOX'
        Process='PROCESS'
        Export='EXPORT'
        Report = 'REPORT'
        VoxelPC = 'VOXPC'

    class lambda_type:
        Generator = 'VOXEL'
        Report = 'REPORT'
        View = 'VIEW'
        Workflow = 'WORKFLOW'
        Process = 'PROCESS'

    CRSFields = [
        "coord_origin_x",
        "coord_origin_y",
        "coord_origin_z",
        "coord_hdatum",
        "coord_vdatum",
        "coord_projection",
        "coord_unit",
        "voxel_size",
        "coord_projection_tm_falseEasting",
        "coord_projection_tm_falseNorthing",
        "coord_projection_tm_latOriginDeg",
        "coord_projection_tm_longMeridianDeg",
        "coord_projection_tm_scaleFactor",
        "coord_projection_lcc_falseEasting",
        "coord_projection_lcc_falseNorthing",
        "coord_projection_lcc_latOfOriginDeg",
        "coord_projection_lcc_longOfOriginDeg", 
        "coord_projection_lcc_firstStdParallelDeg",
        "coord_projection_lcc_secondStdParallelDeg",
        "coord_projection_aeac_falseEasting",
        "coord_projection_aeac_falseNorthing",
        "coord_projection_aeac_latOfOriginDeg",
        "coord_projection_aeac_longOfOriginDeg", 
        "coord_projection_aeac_firstStdParallelDeg",
        "coord_projection_aeac_secondStdParallelDeg",
        "coord_projection_amg_zone",
        "coord_projection_utm_easting",
        "coord_projection_utm_northing"]

    DefaultFields = {
        entity_type.VoxelTerrain : {
            'include_classification' : '0 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15 16 17 18 19 20 21 22 23 24 25 26 27 28 29 30 31',
        },
        entity_type.VoxelMesh : {
            'translate_x' : '0',
            'translate_y' : '0',
            'translate_z' : '0',
            'scale_x' : '1',
            'scale_y' : '1',
            'scale_z' : '1',
            'rotate_x' : '0',
            'rotate_y' : '0',
            'rotate_z' : '0',
            'rotation_order' : 'ZXZ',
        },
        entity_type.IndexedMesh : {
            'translate_x' : '0',
            'translate_y' : '0',
            'translate_z' : '0',
            'scale_x' : '1',
            'scale_y' : '1',
            'scale_z' : '1',
            'rotate_x' : '0',
            'rotate_y' : '0',
            'rotate_z' : '0',
            'rotation_order' : 'ZXZ',
        },
        entity_type.BlockModel : {
            'translate_x' : '0',
            'translate_y' : '0',
            'translate_z' : '0',
            'scale_x' : '1',
            'scale_y' : '1',
            'scale_z' : '1',
            'rotate_x' : '0',
            'rotate_y' : '0',
            'rotate_z' : '0',
            'rotation_order' : 'ZXZ',
            'bm_type': 'VOXBM1'
        },
        entity_type.OrthoVox : {
            'translate_x' : '0',
            'translate_y' : '0',
            'translate_z' : '0',
            'scale_x' : '1',
            'scale_y' : '1',
            'scale_z' : '1',
            'rotate_x' : '0',
            'rotate_y' : '0',
            'rotate_z' : '0',
            'rotation_order' : 'ZXZ'
        }
    }

    HTTPSession = requests.Session()

    def __init__(self, apiurl="https://api.voxelfarm.com"):
        self.proxy = None
        self.VFAPIURL = apiurl
        self.VFAPIURL_ENTITY = self.VFAPIURL + '/entity.ashx'
        self.VFAPIURL_FILE = self.VFAPIURL + '/file.ashx'
        self.VFAPIURL_EVENT = self.VFAPIURL + '/events.ashx'
        self.token: str = None
        self.aad_credentials : dict = None
        self.ts_api_version = "2020-07-31"
        self.ts_environmentName = os.getenv('vf_ts_hub_name')
        self.ts_applicationName = os.getenv('vf_ts_app_name')

    def set_proxy(self, proxy):
        self.proxy = proxy

    class api_result:
        def __init__(self):      
            self.success = False
            self.error_info = ''

    class creation_result(api_result):
        def __init__(self):      
            self.id = ''

    class multiple_creation_result():
        def __init__(self):      
            self.success = False
            self.error_info = ''
            self.ids = []

    class wait_result(api_result):
        def __init__(self):      
            self.complete = False
            self.times = {}

    def add_default_fields(self, type, fields):
        if type in self.DefaultFields:
            for field in self.DefaultFields[type]:
                if not field in fields:
                    fields[field] = self.DefaultFields[type][field]
        if not 'file_date' in fields:
            fields['file_date'] = str(1000 * int(time.time()))

    class crs_result(api_result):
        def __init__(self):      
            self.crs = {}

    def get_project_crs(self, project):
        result = self.crs_result()
        params = {'id': project, 'project': project}

        projectRequest = self.HTTPSession.get(self.VFAPIURL_ENTITY, params=params, headers=self.get_auth_headers(), proxies=self.proxy)
        if projectRequest.status_code != 200:
            result.success = False
            result.error_info = f'HTTP Error code {projectRequest.status_code}' 
            return result
        entityJson = json.loads(projectRequest.text)
        values = {
            'coord_hunit': entityJson['coord_unit'],
            'coord_vunit': entityJson['coord_unit']
        }
        for field in self.CRSFields:
            if field in entityJson:
                values[field] = entityJson[field]
            else:
                values[field] = '0'
        result.crs = values
        result.success = True
        return result

    def get_entity(self, id, project = None):
        params={'id': id}
        if project:
            params['project'] = project
        projectRequest = self.HTTPSession.get(self.VFAPIURL_ENTITY, params=params, headers=self.get_auth_headers(), proxies=self.proxy)
        if projectRequest.status_code != 200:
            return None
        entityJson = json.loads(projectRequest.text)
        return entityJson

    def get_projects_names(self):
        result = self.collection_result()
        entityRequest = self.HTTPSession.get(self.VFAPIURL_ENTITY, headers=self.get_auth_headers())
        if (entityRequest.status_code == 200):
            entityJson = entityRequest.json()
            items = []
            for x in entityJson:
                v = entityJson[x]
                if v["type"] == "PROJECT_REF":
                    items.append(v["project_ref"])

            result.items = items
            result.success = True
        else:
            result.success = False
            result.error_info = f'HTTP Error code {entityRequest.status_code}'  
        return result

    def create_project(self, name, fields):
        result = self.creation_result()
        if fields == None:
            fields = {}
        fields['name'] = name
        self.add_default_fields(type, fields)
        data = {'operation': 'create', 'data': json.dumps(fields)}
        entityRequest = self.HTTPSession.post(self.VFAPIURL_ENTITY, data=data, headers=self.get_auth_headers(), proxies=self.proxy)
        if entityRequest.status_code != 200:
            result.success = False
            result.error_info = f'HTTP Error code {entityRequest.status_code}'
            return result

        apiResult = json.loads(entityRequest.text)
        if ('result' not in apiResult) or (apiResult['result'] != 'success'):
            result.success = False
            if 'error_info' in apiResult:
                result.error_info = base64.b64decode(apiResult['error_info']).decode('ascii')
            else:    
                result.error_info = 'error_info not found'
            return result
        result.success = True
        result.id = apiResult['id']
        return result

    def delete_project(self, project):
        items = []
        # Get all entities in project
        result = self.get_collection(project)
        if not result.success:
            print(result.error_info)
            exit(3)
        entities = result.items
        for e in entities:
            entity = entities[e]
            if entity['type'] != 'USER':
                items.append(entity["ID"])
        # Get all user_ref objects for project (users associated with the project)
        result = self.get_collection('users:' + project)
        if not result.success:
            print(result.error_info)
            exit(4)
        user_refs = result.items
        for user_ref in user_refs:
            if user_refs[user_ref]['type'] == 'USER_REF':
                user_id = user_refs[user_ref]['user_ref']
                items.append(user_id + ':' + project)
                items.append(project + ':' + user_id)
        items.append('R3Vlc3Q=:' + project) # guest user
        self.delete_entities(project, items)

    def clean_project(self, project):
        items = []
        # Get all entities in project
        result = self.get_collection(project)
        if not result.success:
            print(result.error_info)
            exit(3)
        entities = result.items
        for e in entities:
            entity = entities[e]
            if entity['type'] != 'USER' and entity['type'] != 'PROJECT':
                items.append(entity["ID"])

        self.delete_entities(project, items)        

    def create_entity_raw(self, project, type, name, fields, crs):
        result = self.creation_result()
        if fields == None:
            fields = {}
        fields['name'] = name
        fields['type'] = 'FILE'
        fields['file_type'] = type
        fields['state'] = 'COMPLETE'
        fields['project'] = project
        for key, value in crs.items():
            fields['entity_' + key] = value
        self.add_default_fields(type, fields)
        params={'project': project}
        entityRequest = self.HTTPSession.post(self.VFAPIURL_ENTITY, params=params, headers=self.get_auth_headers(), json=fields, proxies=self.proxy)
        if entityRequest.status_code != 200:
            result.success = False
            result.error_info = f'HTTP Error code {entityRequest.status_code}' 
            return result

        apiResult = json.loads(entityRequest.text)
        if ('result' not in apiResult) or (apiResult['result'] != 'success'):
            result.success = False
            if 'error_info' in apiResult:
                result.error_info = base64.b64decode(apiResult['error_info']).decode('ascii')
            else:    
                result.error_info = 'error_info not found'
            return result
        result.success = True
        result.id = apiResult['id']
        return result

    def update_entity(self, project, id, fields):
        result = self.creation_result()
        if fields == None:
            fields = {}
        fields['ID'] = id
        fields['project'] = project
        params={'project': project, 'id': id}
        entityRequest = self.HTTPSession.post(self.VFAPIURL_ENTITY, params=params, headers=self.get_auth_headers(), json=fields, proxies=self.proxy)
        if entityRequest.status_code != 200:
            result.success = False
            result.error_info = f'HTTP Error code {entityRequest.status_code}'
            return result

        apiResult = json.loads(entityRequest.text)
        if ('result' not in apiResult) or (apiResult['result'] != 'success'):
            result.success = False
            if 'error_info' in apiResult:
                result.error_info = base64.b64decode(apiResult['error_info']).decode('ascii')
            else:    
                result.error_info = 'error_info not found'
            return result

        result.success = True
        result.id = id
        return result

    def create_folder(self, project, folder, name, fields = {}):
        result = self.creation_result()
        if folder == None:
            folder = '0'

        fields['name'] = name
        fields['type'] = 'FILE'
        fields['file_type'] = 'FOLDER'
        fields['state'] = 'COMPLETE'
        fields['project'] = project
        fields['file_folder'] = folder
        self.add_default_fields(type, fields)
        params={'project': project}
        entityRequest = self.HTTPSession.post(self.VFAPIURL_ENTITY, params=params, headers=self.get_auth_headers(), json=fields, proxies=self.proxy)
        if entityRequest.status_code != 200:
            result.success = False
            result.error_info = f'HTTP Error code {entityRequest.status_code}'
            return result

        apiResult = json.loads(entityRequest.text)
        if ('result' not in apiResult) or (apiResult['result'] != 'success'):
            result.success = False
            if 'error_info' in apiResult:
                result.error_info = base64.b64decode(apiResult['error_info']).decode('ascii')
            else:    
                result.error_info = 'error_info not found'
            return result

        result.success = True
        result.id = apiResult['id']
        return result

    def create_entity_processed(self, project, type, name, fields, crs, callback=None, update_type = None):
        result = self.creation_result()
        if fields == None:
            fields = {}
        fields['name'] = name
        fields['type'] = 'FILE'
        fields['file_type'] = type
        fields['project'] = project
        fields['state'] = 'PARTIAL'
        if update_type:
            fields['workflow_update_type'] = update_type
        for key, value in crs.items():
            fields[key] = value

        self.add_default_fields(type, fields)
        params={'project': project}
        entityRequest = self.HTTPSession.post(self.VFAPIURL_ENTITY, params=params, headers=self.get_auth_headers(), json=fields, proxies=self.proxy)
        if entityRequest.status_code != 200:
            result.success = False
            result.error_info = f'HTTP Error code {entityRequest.status_code}'
            return result

        apiResult = json.loads(entityRequest.text)
        if ('result' not in apiResult) or (apiResult['result'] != 'success'):
            result.success = False
            if 'error_info' in apiResult:
                result.error_info = base64.b64decode(apiResult['error_info']).decode('ascii')
            else:    
                result.error_info = 'error_info not found'
            return result

        result.id = apiResult['id']
        params={'project': project, 'org': self.OrgId, 'id': result.id}
        data={'process':'PROCESS'}
        if callback != None:
            data['callback'] = callback
        entityRequest = self.HTTPSession.post(self.VFAPIURL_EVENT, params=params, headers=self.get_auth_headers(), data=data, proxies=self.proxy)
        if entityRequest.status_code != 200:
            result.success = False
            result.error_info = f'HTTP Error code {entityRequest.status_code}'
            return result
        apiResult = json.loads(entityRequest.text)
        if ('result' not in apiResult) or (apiResult['result'] != 'success'):
            result.success = False
            if 'error_info' in apiResult:
                result.error_info = base64.b64decode(apiResult['error_info']).decode('ascii')
            else:    
                result.error_info = 'error_info not found'
            return result
        result.success = True
        return result

    def create_process_entity(self, project, type, name, fields, crs, files, callback=None, update_type = None):
        result = self.creation_result()

        # Create the Entity
        if fields == None:
            fields = {}
        fields['name'] = name
        fields['type'] = 'FILE'
        fields['file_type'] = type
        fields['state'] = 'PARTIAL'
        fields['project'] = project
        fields['code_origin'] = 'PYTHON_FILE'
        if update_type:
            fields['workflow_update_type'] = update_type

        for key, value in crs.items():
            fields['entity_' + key] = value
        self.add_default_fields(type, fields)
        params={'project': project}
        entityRequest = self.HTTPSession.post(self.VFAPIURL_ENTITY, params=params, headers=self.get_auth_headers(), json=fields, proxies=self.proxy)
        if entityRequest.status_code != 200:
            result.success = False
            result.error_info = f'HTTP Error code {entityRequest.status_code}'
            return result

        apiResult = json.loads(entityRequest.text)
        if ('result' not in apiResult) or (apiResult['result'] != 'success'):
            result.success = False
            if 'error_info' in apiResult:
                result.error_info = base64.b64decode(apiResult['error_info']).decode('ascii')
            else:    
                result.error_info = 'error_info not found'
            return result

        result.id = apiResult['id']

        # Attach files to the entity
        if files:
            for file in files:
                if (not os.path.exists(file)):
                    result.success = False
                    result.error_info = 'File not found: ' + file
                    return result

                attach_files = {'file': open(file, 'rb')}
                apiResult = self.attach_files(project=project, id=result.id, files=attach_files)
                if not apiResult.success:
                    result.success = False
                    result.error_info = base64.b64decode(apiResult['error_info']).decode('ascii')
                    return result

        fields.clear()
        fields['partial_status'] = ""
        fields['partial_progress'] = "0"

        result = self.update_entity(
            project=project, 
            id=result.id, 
            fields=fields)

        # Process the Entity
        params={'project': project, 'org': self.OrgId, 'id': result.id}
        data={'process':'PROCESS'}
        if callback != None:
            data['callback'] = callback
        entityRequest = self.HTTPSession.post(self.VFAPIURL_EVENT, params=params, headers=self.get_auth_headers(), data=data, proxies=self.proxy)
        if entityRequest.status_code != 200:
            result.success = False
            result.error_info = f'HTTP Error code {entityRequest.status_code}'
            return result

        apiResult = json.loads(entityRequest.text)
        if ('result' not in apiResult) or (apiResult['result'] != 'success'):
            result.success = False
            if 'error_info' in apiResult:
                result.error_info = base64.b64decode(apiResult['error_info']).decode('ascii')
            else:    
                result.error_info = 'error_info not found'
            return result
        result.success = True
        return result        

    def create_python_process_entity(self, project, name, code, fields, inputs, files = None, callback = None, update_type = None):
        result = self.creation_result()

        # Create the Entity
        if fields == None:
            fields = {}
        fields['name'] = name
        fields['type'] = 'FILE'
        fields['file_type'] = 'PROCESS'
        fields['state'] = 'PARTIAL'
        fields['project'] = project

        fields['code'] = code
        if files and code in files:
            fields['code_origin'] = 'PYTHON_FILE'
        else:
            fields['code_origin'] = 'PROGRAM'

        if update_type:
            fields['workflow_update_type'] = update_type

        if inputs:
            for input in inputs:
                fields['input_value_' + input] = inputs[input]
                fields['input_filter_' + input] = '0'
                fields['input_type_' + input] = '0'
                fields['input_label_' + input] = '0'

        self.add_default_fields(type, fields)
        params={'project': project}
        entityRequest = self.HTTPSession.post(self.VFAPIURL_ENTITY, params=params, headers=self.get_auth_headers(), json=fields, proxies=self.proxy)
        if entityRequest.status_code != 200:
            result.success = False
            result.error_info = f'HTTP Error code {entityRequest.status_code}'
            return result

        apiResult = json.loads(entityRequest.text)
        if ('result' not in apiResult) or (apiResult['result'] != 'success'):
            result.success = False
            if 'error_info' in apiResult:
                result.error_info = base64.b64decode(apiResult['error_info']).decode('ascii')
            else:    
                result.error_info = 'error_info not found'
            return result

        result.id = apiResult['id']

        # Attach files to the entity
        if files:
            for file in files:
                if (not os.path.exists(file)):
                    result.success = False
                    result.error_info = 'File not found: ' + file
                    return result

                attach_files = {'file': open(file, 'rb')}
                apiResult = self.attach_files(project=project, id=result.id, files=attach_files)
                if not apiResult.success:
                    result.success = False
                    result.error_info = base64.b64decode(apiResult['error_info']).decode('ascii')
                    return result

        fields.clear()
        fields['partial_status'] = ""
        fields['partial_progress'] = "0"

        result = self.update_entity(
            project=project, 
            id=result.id, 
            fields=fields)

        # Process the Entity
        params={'project': project, 'org': self.OrgId, 'id': result.id}
        data={'process':'PROCESS'}
        if callback != None:
            data['callback'] = callback
        entityRequest = self.HTTPSession.post(self.VFAPIURL_EVENT, params=params, headers=self.get_auth_headers(), data=data, proxies=self.proxy)
        if entityRequest.status_code != 200:
            result.success = False
            result.error_info = f'HTTP Error code {entityRequest.status_code}'
            return result

        apiResult = json.loads(entityRequest.text)
        if ('result' not in apiResult) or (apiResult['result'] != 'success'):
            result.success = False
            if 'error_info' in apiResult:
                result.error_info = base64.b64decode(apiResult['error_info']).decode('ascii')
            else:    
                result.error_info = 'error_info not found'
            return result
        result.success = True
        return result        

    def create_workflow_entity(self, project, name, fields, inputs):
        result = self.creation_result()

        # Create the Entity
        if fields == None:
            fields = {}
        fields['name'] = name
        fields['type'] = 'FILE'
        fields['file_type'] = 'PROCESS'
        fields['process_type'] = 'WORKFLOW'
        fields['state'] = 'PARTIAL'
        fields['project'] = project

        for input in inputs:
            fields['input_value_' + input] = inputs[input]
            fields['input_filter_' + input] = '0'
            fields['input_type_' + input] = '0'
            fields['input_label_' + input] = '0'

        self.add_default_fields(type, fields)
        params={'project': project}
        entityRequest = self.HTTPSession.post(self.VFAPIURL_ENTITY, params=params, headers=self.get_auth_headers(), json=fields, proxies=self.proxy)
        if entityRequest.status_code != 200:
            result.success = False
            result.error_info = f'HTTP Error code {entityRequest.status_code}'
            return result

        apiResult = json.loads(entityRequest.text)
        if ('result' not in apiResult) or (apiResult['result'] != 'success'):
            result.success = False
            if 'error_info' in apiResult:
                result.error_info = base64.b64decode(apiResult['error_info']).decode('ascii')
            else:    
                result.error_info = 'error_info not found'
            return result

        result.id = apiResult['id']

        fields.clear()
        fields['partial_status'] = ""
        fields['partial_progress'] = "0"

        result = self.update_entity(
            project=project, 
            id=result.id, 
            fields=fields)

        # Process the Entity
        params={'project': project, 'org': self.OrgId, 'id': result.id}
        data={'process':'PROCESS'}
        entityRequest = self.HTTPSession.post(self.VFAPIURL_EVENT, params=params, headers=self.get_auth_headers(), data=data, proxies=self.proxy)
        if entityRequest.status_code != 200:
            result.success = False
            result.error_info = f'HTTP Error code {entityRequest.status_code}'
            return result

        apiResult = json.loads(entityRequest.text)
        if ('result' not in apiResult) or (apiResult['result'] != 'success'):
            result.success = False
            if 'error_info' in apiResult:
                result.error_info = base64.b64decode(apiResult['error_info']).decode('ascii')
            else:    
                result.error_info = 'error_info not found'
            return result
        result.success = True
        return result        

    def create_report(self, project, program, lod, region, name, fields, inputs, callback=None, update_type = None):
        result = self.creation_result()
        if fields == None:
            fields = {}
        fields['name'] = name
        fields['type'] = 'FILE'
        fields['file_type'] = 'REPORT'
        fields['program'] = program
        fields['project'] = project
        fields['region'] = region
        fields['lod'] = str(lod)
        fields['state'] = 'PARTIAL'
        if update_type:
            fields['workflow_update_type'] = update_type
        for input in inputs:
            fields['input_value_' + input] = inputs[input]
            fields['input_filter_' + input] = '0'
            fields['input_type_' + input] = '0'
            fields['input_label_' + input] = '0'
        self.add_default_fields(type, fields)
        params={'project': project}
        entityRequest = self.HTTPSession.post(self.VFAPIURL_ENTITY, params=params, headers=self.get_auth_headers(), json=fields, proxies=self.proxy)
        if entityRequest.status_code != 200:
            result.success = False
            result.error_info = f'HTTP Error code {entityRequest.status_code}'
            return result

        apiResult = json.loads(entityRequest.text)
        if ('result' not in apiResult) or (apiResult['result'] != 'success'):
            result.success = False
            if 'error_info' in apiResult:
                result.error_info = base64.b64decode(apiResult['error_info']).decode('ascii')
            else:    
                result.error_info = 'error_info not found'
            return result

        result.id = apiResult['id']
        params={'project': project, 'org': self.OrgId, 'id': result.id}
        data={'process':'RUN_REPORT'}
        if callback != None:
            data['callback'] = callback
        entityRequest = self.HTTPSession.post(self.VFAPIURL_EVENT, params=params, headers=self.get_auth_headers(), data=data, proxies=self.proxy)
        if entityRequest.status_code != 200:
            result.success = False
            result.error_info = f'HTTP Error code {entityRequest.status_code}'
            return result

        apiResult = json.loads(entityRequest.text)
        if ('result' not in apiResult) or (apiResult['result'] != 'success'):
            result.success = False
            if 'error_info' in apiResult:
                result.error_info = base64.b64decode(apiResult['error_info']).decode('ascii')
            else:    
                result.error_info = 'error_info not found'
        else:
            result.success = True

        return result

    def create_export(self, project, program, lod, region, name, fields, inputs, callback=None, update_type = None):
        result = self.creation_result()
        if fields == None:
            fields = {}
        fields['name'] = name
        fields['type'] = 'FILE'
        fields['file_type'] = 'EXPORT'
        fields['program'] = program
        fields['project'] = project
        fields['region'] = region
        fields['lod'] = str(lod)
        fields['state'] = 'PARTIAL'
        if update_type:
            fields['workflow_update_type'] = update_type
        for input in inputs:
            fields['input_value_' + input] = inputs[input]
            fields['input_filter_' + input] = '0'
            fields['input_type_' + input] = '0'
            fields['input_label_' + input] = '0'
        self.add_default_fields(type, fields)
        params={'project': project}
        entityRequest = self.HTTPSession.post(self.VFAPIURL_ENTITY, params=params, headers=self.get_auth_headers(), json=fields, proxies=self.proxy)
        if entityRequest.status_code != 200:
            result.success = False
            result.error_info = f'HTTP Error code {entityRequest.status_code}'
            return result

        apiResult = json.loads(entityRequest.text)
        if ('result' not in apiResult) or (apiResult['result'] != 'success'):
            result.success = False
            if 'error_info' in apiResult:
                result.error_info = base64.b64decode(apiResult['error_info']).decode('ascii')
            else:    
                result.error_info = 'error_info not found'
            return result
        result.id = apiResult['id']
        params={'project': project, 'org': self.OrgId, 'id': result.id}
        data={'process':'RUN_EXPORT'}
        if callback != None:
            data['callback'] = callback
        entityRequest = self.HTTPSession.post(self.VFAPIURL_EVENT, params=params, headers=self.get_auth_headers(), data=data, proxies=self.proxy)
        if entityRequest.status_code != 200:
            result.success = False
            result.error_info = f'HTTP Error code {entityRequest.status_code}'
            return result

        apiResult = json.loads(entityRequest.text)
        if ('result' not in apiResult) or (apiResult['result'] != 'success'):
            result.success = False
            if 'error_info' in apiResult:
                result.error_info = base64.b64decode(apiResult['error_info']).decode('ascii')
            else:    
                result.error_info = 'error_info not found'
        else:
            result.success = True

        return result

    def reprocess_entity(self, id):
        result = self.creation_result()
        result.id = id

        report_entity = self.get_entity(id)
        if report_entity:
            project = report_entity["project"]

            params = {
                'project': project,
                'id' : id
            }

            fields = {
                'state': 'PARTIAL',
                'partial_progress': '0',
                'partial_status': '',
                'proccessing_time': '0'
            }

            entityRequest = self.HTTPSession.post(self.VFAPIURL_ENTITY, params=params, headers=self.get_auth_headers(), json=fields, proxies=self.proxy)
            if entityRequest.status_code != 200:
                result.success = False
                result.error_info = f'HTTP Error code {entityRequest.status_code}'
                return result

            apiResult = json.loads(entityRequest.text)
            if ('result' not in apiResult) or (apiResult['result'] != 'success'):
                result.success = False
                if 'error_info' in apiResult:
                    result.error_info = base64.b64decode(apiResult['error_info']).decode('ascii')
                else:    
                    result.error_info = 'error_info not found'
                return result

            params={'project': project, 'org': self.OrgId, 'id': id}
            data={'process':'RUN_REPORT'}
            entityRequest = self.HTTPSession.post(self.VFAPIURL_EVENT, params=params, headers=self.get_auth_headers(), data=data, proxies=self.proxy)
            if entityRequest.status_code != 200:
                result.success = False
                result.error_info = f'HTTP Error code {entityRequest.status_code}'
                return result
            apiResult = json.loads(entityRequest.text)
            if ('result' not in apiResult) or (apiResult['result'] != 'success'):
                result.success = False
                if 'error_info' in apiResult:
                    result.error_info = base64.b64decode(apiResult['error_info']).decode('ascii')
                else:    
                    result.error_info = 'error_info not found'
            else:
                result.success = True
        else:
            result.success = False

        return result        

    def create_lambda_python(self, project, type, name, code, fields, files=None):
        result = self.creation_result()
        if fields == None:
            fields = {}

        fields['name'] = name
        fields['type'] = 'FILE'
        fields['file_type'] = 'PROGRAM'
        fields['program_type'] = type
        fields['state'] = 'COMPLETE'

        if code.endswith('.py'):
            fields['code_origin'] = 'PYTHON_FILE'
            fields['code'] = code
        else:
            generatorLambdaCode = code.encode('ascii')
            generatorLambdaCode = base64.b64encode(generatorLambdaCode).decode('ascii')
            fields['code'] = generatorLambdaCode

        fields['project'] = project
        self.add_default_fields(type, fields)
        params={'project': project}
        entityRequest = requests.post(self.VFAPIURL_ENTITY, params=params, headers=self.get_auth_headers(), json=fields, proxies=self.proxy)
        apiResult = json.loads(entityRequest.text)
        if ('result' not in apiResult) or (apiResult['result'] != 'success'):
            result.success = False
            if 'error_info' in apiResult:
                result.error_info = base64.b64decode(apiResult['error_info']).decode('ascii')
            else:    
                result.error_info = 'error_info not found'
            return result
        
        lambda_id = apiResult['id']
        if files:
            for file in files:
                if (not os.path.exists(file)):
                    result.success = False
                    result.error_info = 'File not found: ' + file
                    return result

                attach_files = {'file': open(file, 'rb')}
                apiResult = self.attach_files(project=project, id=lambda_id, files=attach_files)
                if not apiResult.success:
                    result.success = False
                    result.error_info = base64.b64decode(apiResult['error_info']).decode('ascii')
                    return result
                        
        result.success = True
        result.id = lambda_id
        return result
    
    def create_generator(self, project, program, name, fields, inputs):
        result = self.creation_result()
        if fields == None:
            fields = {}
        fields['name'] = name
        fields['type'] = 'FILE'
        fields['file_type'] = 'VOXGEN'
        fields['program'] = program
        fields['state'] = 'COMPLETE'
        fields['project'] = project
        for input in inputs:
            fields['input_value_' + input] = inputs[input]
            fields['input_filter_' + input] = '0'
            fields['input_type_' + input] = '0'
            fields['input_label_' + input] = '0'
        self.add_default_fields(type, fields)
        params={'project': project}
        entityRequest = requests.post(self.VFAPIURL_ENTITY, params=params, headers=self.get_auth_headers(), json=fields, proxies=self.proxy)
        apiResult = json.loads(entityRequest.text)
        if ('result' not in apiResult) or (apiResult['result'] != 'success'):
            result.success = False
            if 'error_info' in apiResult:
                result.error_info = base64.b64decode(apiResult['error_info']).decode('ascii')
            else:    
                result.error_info = 'error_info not found'
            return result
        result.success = True
        result.id = apiResult['id']
        return result

    def is_processing_complete(self, project, ids):
        result = self.wait_result()
        for id in ids:
            params = {'id': id, 'project': project}
            entityRequest = self.HTTPSession.get(self.VFAPIURL_ENTITY, params=params, headers=self.get_auth_headers())
            if entityRequest.status_code == 200:
                entityJson = json.loads(entityRequest.text)
                if not 'state' in entityJson:
                    result.complete = False
                    result.success = False
                    result.error_info = json.dumps(entityJson)
                    break
                if entityJson['state'] == 'PARTIAL':
                    result.complete = False
                    result.success = True
                    break
                if entityJson['state'] == 'COMPLETE':
                    result.complete = True
                    result.success = True
                    if 'processing_time' in entityJson:
                        result.times[id] = entityJson['processing_time']
                    else:    
                        result.times[id] = ''
                if entityJson['state'] == 'ERROR':
                    result.complete = False
                    result.success = False
                    result.error_info = base64.b64decode(entityJson['error_info']).decode('ascii')
                    break
            else:
                result.complete = False
                result.success = False
                result.error_info = 'HTTP Error ' + str(entityRequest.status_code)
                break

        return result

    def delete_entities(self, project, ids):
        result = self.api_result()
        deleteIdentifierList = ''
        for id in ids:
            deleteIdentifierList = deleteIdentifierList + id + ' '
        params = {'project': project, 'org': self.OrgId, 'operation': 'delete'}
        data = {'operation': 'delete', 'items': deleteIdentifierList}
        entityRequest = self.HTTPSession.post(self.VFAPIURL_ENTITY, params=params, headers=self.get_auth_headers(), data=data, proxies=self.proxy)
        if entityRequest.status_code != 200:
            result.success = False
            result.error_info = 'Error deleting entities'
            return result

        apiResult = json.loads(entityRequest.text)
        if ('result' not in apiResult) or (apiResult['result'] != 'success'):
            result.success = False
            if 'error_info' in apiResult:
                result.error_info = base64.b64decode(apiResult['error_info']).decode('ascii')
            else:    
                result.error_info = 'error_info not found'
            return result
        else:    
            result.success = True
        return result

    def attach_files(self, project, id, files):
        result = self.api_result()
        params = {'project': project, 'org': self.OrgId, 'id': id}
        entityRequest = self.HTTPSession.post(self.VFAPIURL_FILE, params=params, headers=self.get_auth_headers(), files=files, proxies=self.proxy)
        if entityRequest.status_code != 200:
            result.success = False
            result.error_info = f'HTTP Error code {entityRequest.status_code}'
            return result
        apiResult = json.loads(entityRequest.text)
        if ('result' not in apiResult) or (apiResult['result'] != 'success'):
            result.success = False
            if 'error_info' in apiResult:
                result.error_info = base64.b64decode(apiResult['error_info']).decode('ascii')
            else:    
                result.error_info = 'error_info not found'
            return result
        else:    
            result.success = True
        return result

    def get_file(self, project, entity, filename):
        params = {'id': entity, 'project': project, 'org': self.OrgId, 'filename': filename}
        entityRequest = self.HTTPSession.get(self.VFAPIURL_FILE, params=params, headers=self.get_auth_headers(), proxies=self.proxy)
        if entityRequest.status_code != 200:
            return ''
        return entityRequest.text

    def get_binary_file(self, project, entity, filename):
        params = {'id': entity, 'project': project, 'org': self.OrgId, 'filename': filename}
        entityRequest = self.HTTPSession.get(self.VFAPIURL_FILE, params=params, headers=self.get_auth_headers(), proxies=self.proxy)
        if entityRequest.status_code != 200:
            return None
        return entityRequest.content  

    def download_file(self, project, entity, filename, destination):
        params = {'id': entity, 'project': project, 'org': self.OrgId, 'filename': filename}
        entityRequest = self.HTTPSession.get(self.VFAPIURL_FILE, params=params, headers=self.get_auth_headers(), proxies=self.proxy)
        if (entityRequest.status_code) != 200:
            return False
        with open(destination, 'wb') as file:
            file.write(entityRequest.content)
        return True

    def get_file_list(self, project, entity):
        params = {'id': entity, 'project': project, 'org': self.OrgId}
        entityRequest = self.HTTPSession.get(self.VFAPIURL_FILE, params=params, headers=self.get_auth_headers(), proxies=self.proxy)

        list = []
        if entityRequest.status_code == 200:
            for line in entityRequest.text.split("\n"):
                if not line.strip():
                    continue
                list.append(line.lstrip())        
        return list

    class collection_result(api_result):
        items = None

    def get_collection(self, id):
        result = self.collection_result()
        params = {'project': id}
        entityRequest = self.HTTPSession.get(self.VFAPIURL_ENTITY, params=params, headers=self.get_auth_headers())
        if (entityRequest.status_code == 200):
            entityJson = json.loads(entityRequest.text)
            result.items = entityJson
            result.success = True
        else:
            result.success = False
            result.error_info = f'HTTP Error code {entityRequest.status_code}'
        return result

    def compute_region_from_aabbs(self, entity_id):
        entity = self.get_entity(entity_id)
        min_x = float(entity["aabb_min_x"]) - 20
        max_x = float(entity["aabb_max_x"]) + 20
        min_y = float(entity["aabb_min_y"]) - 20
        max_y = float(entity["aabb_max_y"]) + 20
        min_z = float(entity["aabb_min_z"]) - 20
        max_z = float(entity["aabb_max_z"]) + 20
        region = '1'
        region += "," + str(min_z)
        region += "," + str(max_z)
        region += "," + str(min_x)
        region += "," + str(min_y)
        region += "," + str(max_x)
        region += "," + str(min_y)
        region += "," + str(max_x)
        region += "," + str(max_y)
        region += "," + str(min_x)
        region += "," + str(max_y)
        return region

    def workflow_get_product_property(self, project_id, product_id, property, version = 0):
        project = self.get_entity(project_id, project_id)
        if project:
            product_key = 'workflow_folder_' + product_id
            if product_key in project:
                product_folder_id = project[product_key]
                if version > 0:
                    apiResult = self.get_collection(project_id)
                    if apiResult.success:
                        version_name = 'Version ' + str(version)
                        entities = apiResult.items
                        for e in entities:
                            entity = entities[e]
                            if ('file_folder' in entity) and (entity['file_folder'] == product_folder_id) and ('file_type' in entity) and (entity['file_type'] == 'FOLDER'):
                                if entity["name"] == version_name:
                                    extended_prop = 'version_' + property
                                    if extended_prop in entity:
                                        return entity[extended_prop]
                                    break    
                else:
                    product_entity = self.get_entity(product_folder_id, project_id)
                    if product_entity and ('version_active' in product_entity):
                        active_version_id = product_entity['version_active']
                        active_version = self.get_entity(active_version_id, project_id)
                        if active_version:
                            extended_prop = 'version_' + property
                            if extended_prop in active_version:
                                return active_version[extended_prop]
        return None

    def workflow_get_parameter_dataframe(self, project_id, product_id, version = 0):
        attribute_product = self.workflow_get_product_property(project_id, product_id, 'report_entity', version)
        if attribute_product:
            attribute = self.get_file(project_id, attribute_product, 'report.csv')
            if attribute:
                content = StringIO(attribute)
                return pd.read_csv(content)
        return pd.DataFrame()

    def workflow_get_entity_id(self, project_id, product_id, alias_id, version = 0):
        project = self.get_entity(project_id, project_id)
        if project:
            product_key = 'workflow_folder_' + product_id
            if product_key in project:
                product_folder_id = project[product_key]
                if version > 0:
                    apiResult = self.get_collection(project_id)
                    if apiResult.success:
                        version_name = 'Version ' + str(version)
                        entities = apiResult.items
                        for e in entities:
                            entity = entities[e]
                            if ('file_folder' in entity) and (entity['file_folder'] == product_folder_id) and ('file_type' in entity) and (entity['file_type'] == 'FOLDER'):
                                if entity["name"] == version_name:
                                    extended_prop = 'alias_' + alias_id
                                    if extended_prop in entity:
                                        return entity[extended_prop]
                                    break    
                else:
                    product_entity = self.get_entity(product_folder_id, project_id)
                    if product_entity and ('version_active' in product_entity):
                        active_version_id = product_entity['version_active']
                        active_version = self.get_entity(active_version_id, project_id)
                        if active_version:
                            extended_prop = 'alias_' + alias_id
                            if extended_prop in active_version:
                                return active_version[extended_prop]
        return None

    def workflow_get_dataframe(self, project_id, product_id, alias_id, file_name = None, version = 0):
        attribute_product = self.workflow_get_entity_id(project_id, product_id, alias_id, version)
        if attribute_product:
            dataframe_file = 'report.csv'
            if file_name:
                dataframe_file = file_name

            split_tup = os.path.splitext(dataframe_file)
            file_extension = split_tup[1].lower()
            if file_extension == '.csv':
                attribute = self.get_file(project_id, attribute_product, dataframe_file)
                if attribute:
                    content = StringIO(attribute)
                    return pd.read_csv(content)
            elif file_extension == '.ftr':
                attribute = self.get_binary_file(project_id, attribute_product, dataframe_file)
                if attribute:
                    content = io.BytesIO(attribute)
                    return pd.read_feather(content)

        return pd.DataFrame()

    def workflow_get_timings(self, project_id):
        entities = []
        collections = self.get_collection(project_id).items
        
        for entity_id in collections:
            if "processing_start_date" in collections[entity_id] and "file_folder" in collections[entity_id]:
                file_folder_entity = collections[collections[entity_id]["file_folder"]]
                processing_time = float(collections[entity_id]["processing_time"]) if "processing_time" in collections[entity_id] else 0
                if processing_time > 0:
                    file_date = int(collections[entity_id]["file_date"]) if "file_date" in collections[entity_id] else 0
                    processing_start_date = int(collections[entity_id]["processing_start_date"]) if "processing_start_date" in collections[entity_id] else 0
                    queue_time = processing_start_date - file_date
                    workflow_time = file_date - int(file_folder_entity["capture_date"] if "capture_date" in file_folder_entity else file_date)
                    entities.append({
                        "id": collections[entity_id]["ID"],
                        "file_folder": collections[entity_id]["file_folder"],
                        "file_date": pd.to_datetime(file_date / 1000, unit='s'),
                        "workflow_name": file_folder_entity["workflow_product"] if "workflow_product" in file_folder_entity else "N/A",
                        "name": collections[entity_id]["name"],
                        "state": collections[entity_id]["state"],
                        "version": int(file_folder_entity["version_number"]) if "version_number" in file_folder_entity else "N/A",
                        "workflow_time_ms": workflow_time,
                        "processing_time_ms": processing_time,
                        "queue_time_ms": queue_time
                    })

        return pd.DataFrame.from_dict(entities)    

#region AAD_Credentials
    def set_file_credentials(self, filename: str):
        self.aad_credentials = self.get_env_file_as_dict(filename)
        self.token = self.get_token()
        if (self.token == None):
            print("credentials file: "+filename )
            raise Exception("Check that the credentials are correct")

    def set_credentials(self, tenant: str, client_id: str, client_secret: str):
        self.aad_credentials = {
            "TENANT": tenant,
            "CLIENT_ID": client_id,
            "CLIENT_SECRET":client_secret
        }
        self.token = self.get_token()
        if (self.token == None):
            print("Set Credentials failed ClientId : "+ client_id)
            raise Exception("Check that the credentials are correct")

    def get_env_file_as_dict(self, path: str) -> dict:
        if not os.path.isfile(path):
            t_path = os.path.dirname(__file__)+ path
            if os.path.isfile(t_path):    
                path = t_path
        with open(path, 'r') as f:
            return dict(tuple(line.replace('\n', '').split('=')) for line 
                        in f.readlines() if not line.startswith('#'))

    def get_token(self, ressource=None):
        if ressource is None:
            ressource = self.aad_credentials["CLIENT_ID"]

        if (self.aad_credentials["TENANT"] != None):
            authority = "https://login.microsoftonline.com/"+self.aad_credentials["TENANT"]+"/oauth2/token"
            headers = {'Content-Type': 'application/x-www-form-urlencoded'}
            data = {
                "grant_type" : "client_credentials",
                "client_id" : self.aad_credentials["CLIENT_ID"],
                "client_secret": self.aad_credentials["CLIENT_SECRET"],
                "scope" : "https://graph.microsoft.com/.default",
                "resource": ressource,
            }

            HTTPSession = requests.Session()
            tokenRequest = HTTPSession.post(authority, data=data, headers=headers)
            HTTPSession.close()

            if tokenRequest.status_code == 200:
                tokenResult = tokenRequest.json()
                return tokenResult["access_token"]
        return None

    def check_ExpiredToken(self, token: str):
        body = token.split(".")[1]
        body += "=" * ((4 - len(body) % 4) % 4) #ugh
        txt = base64.b64decode(body).decode('utf-8')
        jwt = json.loads(txt)
        expire = int(jwt["exp"])
        return (datetime.fromtimestamp(expire)) < (datetime.now()+timedelta(minutes=2))

    def get_auth_headers(self):
        if (self.token==None):
            return None
        if self.check_ExpiredToken(self.token):
            self.token = self.get_token()    
        
        return { 'Authorization': f'Bearer {self.token}' }
# endregion AAD_Credentials

# region GetTimeSeriesData
    def get_time_series_data_by_search(self, timespan, projectID, spatialID=None, objectID=None, properties=None, filter=None):
        authorizationToken = self.get_token(
            ressource="https://api.timeseries.azure.com")
        searchString = ""
        if spatialID is not None and spatialID != "" and spatialID != "*":
            searchString = searchString + spatialID + ", "
        if objectID is not None and objectID != "" and objectID != "*":
            searchString = searchString + objectID

        if(searchString[len(searchString)-2:] == ", "):
            searchString = searchString[:len(searchString)-2]

        url = (
            "https://"
            + self.get_environment_id(authorizationToken)
            + ".env.timeseries.azure.com/timeseries/instances/search?"
        )
        querystring = {"api-version": self.ts_api_version}

        payload = {"searchString": f"\"{searchString}\"",
                   "instances": {"pageSize": 100}}

        headers = {
            "Authorization": f"Bearer {authorizationToken}",
            "Content-Type": "application/json",
            "cache-control": "no-cache",
        }

        try:
            jsonResponse = requests.request(
                "POST",
                url,
                data=json.dumps(payload),
                headers=headers,
                params=querystring,
            )
            jsonResponse.raise_for_status()
        except requests.exceptions.ConnectTimeout:
            raise Exception("The request to the TSI api timed out.")
        except requests.exceptions.HTTPError:
            raise Exception(
                "The request to the TSI api returned an unsuccessfull status code.")

        response = json.loads(jsonResponse.text)
        if response["instances"]["hits"]:
            result = response["instances"]["hits"]
            processes = []
            with ThreadPoolExecutor(max_workers=100) as executor:
                for i, _ in enumerate(result):
                    processes.append(
                        executor.submit(
                            lambda p: self.get_data_by_id(*p),
                            [
                                result[i]["timeSeriesId"],
                                properties,
                                filter,
                                timespan,
                                authorizationToken,
                            ],
                        )
                    )

                while "continuationToken" in list(response["instances"].keys()):
                    headers["x-ms-continuation"] = response["instances"]["continuationToken"]
                    jsonResponse = requests.request(
                        "POST",
                        url,
                        data=json.dumps(payload),
                        headers=headers,
                        params=querystring,
                    )
                    jsonResponse.raise_for_status()

                    response = json.loads(jsonResponse.text)
                    result = response["instances"]["hits"]
                    for i, _ in enumerate(result):
                        processes.append(
                            executor.submit(
                                lambda p: self.get_data_by_id(*p),
                                [
                                    result[i]["timeSeriesId"],
                                    properties,
                                    filter,
                                    timespan,
                                    authorizationToken,
                                ],
                            )
                        )
            return self.build_dataframe_from_processes(processes, properties)
        else:
            return pd.DataFrame()

    def get_time_series_data_by_ids(self, projectID, timeseriesIDs, timespan):
        authorizationToken = self.get_token(
            ressource="https://api.timeseries.azure.com")

        processes = []
        with ThreadPoolExecutor(max_workers=100) as executor:
            for i, id in enumerate(timeseriesIDs):
                processes.append(
                    executor.submit(
                        lambda p: self.get_data_by_id(*p),
                        [
                            [id["spatialID"], id["objectID"]],
                            id["properties"] if "properties" in id else None,
                            id["filter"] if "filter" in id else None,
                            timespan,
                            authorizationToken,
                        ],
                    )
                )

        return self.build_dataframe_from_processes(processes, None)

    def get_environment_id(self, authorizationToken):
        url = "https://api.timeseries.azure.com/environments"

        querystring = {"api-version": self.ts_api_version}

        payload = ""
        headers = {
            "x-ms-client-application-name": self.ts_applicationName,
            "Authorization": f"Bearer {authorizationToken}",
            "Content-Type": "application/json",
            "cache-control": "no-cache",
        }

        try:
            response = requests.request(
                "GET",
                url,
                data=payload,
                headers=headers,
                params=querystring,
                timeout=10,
            )
            response.raise_for_status()
        except requests.exceptions.ConnectTimeout:
            raise Exception("The request to the TSI api timed out.")
        except requests.exceptions.HTTPError:
            raise Exception(
                "The request to the TSI api returned an unsuccessfull status code.")

        environments = json.loads(response.text)["environments"]
        environmentId = None
        for environment in environments:
            if environment["displayName"] == self.ts_environmentName:
                environmentId = environment["environmentId"]
                break
        if environmentId == None:
            raise Exception(
                "TSI environment not found. Check the spelling or create an environment in Azure TSI.")

        return environmentId

    def get_data_by_id(
        self,
        timeseries,
        properties,
        filter,
        timespan,
        authorizationToken,
    ):

        url = (
            "https://"
            + self.get_environment_id(authorizationToken)
            + ".env.timeseries.azure.com/timeseries/query?"
        )
        querystring = {"api-version": self.ts_api_version}
        requestType = "getEvents"

        result = {}

        projectedProperties = None
        if(properties != "*" and properties != None and properties != ["*"]):
            projectedProperties = properties
        retry = 0

        payload = {
            requestType: {
                "timeSeriesId": timeseries,
                "searchSpan": {"from": timespan[0], "to": timespan[1]},
                "take": 250000,
                "projectedProperties": projectedProperties,
            }
        }

        if filter is not None and filter != "":
            payload[requestType]["filter"] = {
                "tsx": filter
            }

        headers = {
            "x-ms-client-application-name": self.ts_applicationName,
            "Authorization": f"Bearer {authorizationToken}",
            "Content-Type": "application/json",
            "cache-control": "no-cache",
        }

        while retry < 50:  # Looping with a small random wait if the API returns us problems. This solve an issue where consicutive requests goes beyond TSI max number of concurrent requests.
            try:
                jsonResponse = requests.request(
                    "POST",
                    url,
                    data=json.dumps(payload),
                    headers=headers,
                    params=querystring,
                )
                jsonResponse.raise_for_status()
                break
            except requests.exceptions.HTTPError:
                error = json.loads(jsonResponse.text)
                if(error["error"]["code"] == "TooManyRequests"):
                    # Too many requests at the same time. Waiting for a bit before retrying."
                    time.sleep(random.randint(1, 5))
                    retry += 1
                    continue
                else:
                    raise Exception(
                        "The request to the TSI api returned an unsuccessfull status code.")

        response = json.loads(jsonResponse.text)
        result = response
        result["timeseries"] = timeseries
        retry = 0

        while "continuationToken" in list(response.keys()):
            headers = {
                "x-ms-client-application-name": self.ts_applicationName,
                "Authorization": f"Bearer {authorizationToken}",
                "Content-Type": "application/json",
                "cache-control": "no-cache",
                "x-ms-continuation": response["continuationToken"],
            }
            while retry < 50:
                try:
                    jsonResponse = requests.request(
                        "POST",
                        url,
                        data=json.dumps(payload),
                        headers=headers,
                        params=querystring,
                    )
                    jsonResponse.raise_for_status()
                    break
                except requests.exceptions.HTTPError:
                    error = json.loads(jsonResponse.text)
                    if(error["error"]["code"] == "TooManyRequests"):
                        # Too many requests at the same time. Waiting for a bit before retrying."
                        time.sleep(random.randint(1, 5))
                        retry += 1
                        continue
                    else:
                        raise Exception(
                            "The request to the TSI api returned an unsuccessfull status code.")

            response = json.loads(jsonResponse.text)
            result["timestamps"].extend(response["timestamps"])
            for i, property in enumerate(response["properties"]):
                result["properties"][i]["values"].extend(property["values"])
        retry = 0
        return result

    def build_dataframe_from_processes(self, processes, dTypes):
        if processes is None or len(processes) == 0:
            return pd.DataFrame()

        lambdaData = []
        dfColumns = ["timestamp", "spatialID", "objectID"]
        for task in as_completed(processes):
            result = task.result()
            if "timestamps" in result:
                for i, (timestamp) in enumerate(result["timestamps"]):
                    lambdaData.append({ "timestamp": timestamp,
                                        "spatialID": result["timeseries"][0],
                                        "objectID": result["timeseries"][1],
                                        "properties": []})
                    properties = []
                    for (j, property) in enumerate(result["properties"]):
                        if property["name"] == "spatialId" or property["name"] == "objectId":
                            continue
                        properties.append({
                            "name": property["name"],
                            "value": property["values"][i],
                        })
                        if not property["name"] in dfColumns:
                            dfColumns.append(property["name"])
                    lambdaData[len(lambdaData)-1]["properties"] = properties

        dataframe = pd.DataFrame(columns=dfColumns)
        arDf = []
        for i, (event) in enumerate(lambdaData):
            values = [pd.to_datetime(event["timestamp"]), event["spatialID"], event["objectID"]]
            indexes = ['timestamp', 'spatialID', 'objectID']
            for _, (property) in enumerate(event["properties"]):
                if(property["name"] in indexes):
                    # Received multiple different values for the exact same date and property. Update the value to the latest data we received.
                    values[indexes.index(property["name"])] = property["value"]
                else :
                    values.append(property["value"])
                    indexes.append(property["name"])
            arDf.append(pd.DataFrame([values], columns=indexes))
        dataframe = pd.concat([dataframe, pd.concat(arDf)])

        # Force dtype to what has been requestion in the properties object when calling a search, if any.
        for dtype in enumerate(dTypes):
            name = dtype[1]["name"]
            type = dtype[1]["type"]
            if name in dataframe.columns and type.lower() == 'double':
                dataframe[name] = dataframe[name].astype('float64')
        return dataframe
# endregion GetTimeSeriesData



import requests
import os
import os.path
import datetime
import time
import csv 
import pandas as pd
import json
import voxelfarm
from voxelfarm import voxelfarmclient

class request:
    def __init__(self, framework):
        framework.log('Creating request...')

        vf_api_url = framework.get_vf_api()
        self.vf_api = voxelfarmclient.rest(vf_api_url)
        framework.log(f'vf_api url {vf_api_url}')

        self.lambda_framework = framework

        entity = framework.get_entity()
        self.entity_id = entity['ID']

        framework.log(f'entity_id {self.entity_id}')

        self.scrap_folder = framework.get_scrap_folder()          
        self.raw_entity_id = framework.input_string('raw_entity_id', 'Entity Id', '')
        self.project_id = framework.input_string('project_id', 'Project Id', '')
        self.product_id = framework.input_string('product_id', 'Product Id', '')
        self.product_folder_id = framework.input_string('product_folder_id', 'Product Folder', '')
        self.version_folder_id = framework.input_string('version_folder_id', 'Version Folder', '')
        self.user = framework.input_string('user', 'User', '')
        self.comment = framework.input_string('comment', 'Comment', '')
        self.capture_date = framework.input_string('capture_date', 'Capture Date', '')

        self.update_type = framework.get_callback_update_type()
        self.caller_entity_id = framework.get_callback_entity_id()
        self.caller_entity_state = framework.get_callback_entity_state()

        framework.log(f'scrap_folder {self.scrap_folder}')
        framework.log(f'raw_entity_id {self.raw_entity_id}')
        framework.log(f'project_id {self.project_id}')
        framework.log(f'product_id {self.product_id}')
        framework.log(f'product_folder_id {self.product_folder_id}')
        framework.log(f'version_folder_id {self.version_folder_id}')
        framework.log(f'user {self.user}')
        framework.log(f'comment {self.comment}')
        framework.log(f'capture_date {self.capture_date}')
        framework.log(f'update_type {self.update_type}')
        framework.log(f'caller_entity_id {self.caller_entity_id}')
        framework.log(f'caller_entity_state {self.caller_entity_state}')

        self.crs = {}
        # Get the coordinate system of the project
        framework.log('Retrieving project CRS...')
        result = self.vf_api.get_project_crs(self.project_id)
        if result.success:
            self.crs = result.crs
            framework.log('Retrieved project CRS')

        product_entity = self.vf_api.get_entity(self.product_folder_id, self.project_id)
        framework.log(f'product_entity:{product_entity}')

        # Load previously active version properties
        framework.log('Loading properties from active version...')

        self.active_version_folder_id = "0"
        self.active_version_properties = {}
        if (product_entity and 'version_active' in product_entity and product_entity['version_active'] != '0'):
            self.active_version_folder_id = product_entity['version_active']
            active_version_entity = self.vf_api.get_entity(self.active_version_folder_id, self.project_id)
            if active_version_entity:
                for prop in active_version_entity:
                    if prop.find('version_') == 0:
                        value = active_version_entity[prop]
                        prop_name = prop.replace('version_', '', 1)
                        self.active_version_properties[prop_name] = value

        # Load version properties
        framework.log('Loading properties from current version...')
        self.properties = {}
        version_entity = self.vf_api.get_entity(self.version_folder_id, self.project_id)
        if version_entity:
            for prop in version_entity:
                if prop.find('version_') == 0:
                    value = version_entity[prop]
                    prop_name = prop.replace('version_', '', 1)
                    self.properties[prop_name] = version_entity[prop]

    def get_client_api(self):
        return self.vf_api        

    def get_product_property(self, product_id, property):
        self.lambda_framework.log(f'getting {product_id} property:{property}...')

        workflow_entity = self.vf_api.get_entity(self.project_id, self.project_id)

        field_id = f'workflow_folder_{product_id}'
        if workflow_entity and (field_id in workflow_entity):
            product_folder_id = workflow_entity[field_id]
            self.lambda_framework.log(f'product_folder_id:{product_folder_id}')

            product_entity = self.vf_api.get_entity(product_folder_id, self.project_id)
            if product_entity:
                self.lambda_framework.log(f'product_entity:{product_entity}')
                active_version_id = product_entity['version_active']
                active_version = self.vf_api.get_entity(active_version_id, self.project_id)
                self.lambda_framework.log(f'active_version:{active_version}')
                extended_prop = 'version_' + property
                self.lambda_framework.log(f'extended_prop:{extended_prop}')
                if active_version != None and extended_prop in active_version:
                    return active_version[extended_prop]

        return None

    def get_product_singleton(self, product_id, singleton_id):
        self.lambda_framework.log(f'getting {singleton_id} for {product_id}...')

        workflow_entity = self.vf_api.get_entity(self.project_id, self.project_id)

        field_id = f'workflow_folder_{product_id}'
        if workflow_entity and (field_id in workflow_entity):
            product_folder_id = workflow_entity[field_id]
            self.lambda_framework.log(f'product_folder_id:{product_folder_id}')
            
            product_entity = self.vf_api.get_entity(product_folder_id, self.project_id)
            if product_entity:
                self.lambda_framework.log(f'product_entity: {product_entity}')
                singleton_property = f'workflow_singleton_{singleton_id}'
                if product_entity and singleton_property in product_entity:
                    return product_entity[singleton_property]

        return None

    def define_product_alias(self, alias, value):
        self.lambda_framework.log(f'Define alias {alias} for product {self.product_id}')
        updatedProperties = {}
        alias_property_id = 'alias_' + alias
        updatedProperties[alias_property_id] = value

        self.vf_api.update_entity(
            project=self.project_id,
            id=self.version_folder_id,
            fields=updatedProperties
        )

    def get_callback(self, update_type : str):
        if update_type:
            return f'{self.entity_id}/{update_type}'
        
        return None

class workflow_lambda_framework:

    def __init__(self):  
        pass

    def input_string(self, id, label, default = ""):
        return ""

    def log(self, message):
        pass

    def progress(self, progress, message):
        pass

    def get_entity(self, id = None):
        return None

    def download_entity_file(self, filename, id = None):
        return ""
    
    def get_scrap_folder(self):
        return ""

    def get_tools_folder(self):
        return ""

    def get_entity_folder(self, id = None):
        return ""

    def download_entity_files(self, id = None):
        return ""

    def download_entity_file(self, filename, id = None):
        return ""

    def attach_file(self, filename, id = None):
        pass

    def attach_folder(self, folder, id = None):
        pass

    def upload(self, filename, name, id = None):
        pass

    def set_exit_code(self, code):
        pass

    def get_entity(self, id = None):
        return None

    def get_entity_file_list(self, id = None):
        return []
    
    def get_callback_update_type(self):
        return ""

    def get_callback_entity_id(self):
        return ""

    def get_callback_entity_state(self):
        return ''
    
    def get_vf_api(self):
        return 'http://localhost'
    
    def increment_counter(self, counter, offset):
        return 0
    
    def decrement_counter(self, counter, offset):
        return 0

    def set_counter(self, counter, value):
        pass

    def get_working_dir(self):
        return ''

    def get_notify_emails(self):
        return ''

    def get_notify_users_url(self):
        return ''

    def get_workflow_api(self):
        return ''

    def publish_feed(self, entity_id, timestamp, timestamp_id, spatialId, objectId, defaultObjectId, propertyIds, constant_spatial_id, constant_object_id):
        return {'success': True}

class workflow_lambda_host:

    def __init__(self, framework = None):  
        if framework:
            self.lambda_framework = framework
        else:
            if voxelfarm.voxelfarm_framework:
                self.lambda_framework = voxelfarm.voxelfarm_framework
            else:
                self.lambda_framework = workflow_lambda_framework()

        vf_api_url = self.lambda_framework.get_vf_api()
        self.vf_api = voxelfarmclient.rest(vf_api_url)

    def input_string(self, id, label, default = ""):
        return self.lambda_framework.input_string(id, label, default)

    def log(self, message):
        self.lambda_framework.log(message)

    def progress(self, progress, message):
        self.lambda_framework.progress(progress, message)

    def get_scrap_folder(self):
        return self.lambda_framework.get_scrap_folder()

    def get_tools_folder(self):
        return self.lambda_framework.get_tools_folder()

    def get_entity_folder(self, id = None):
        return self.lambda_framework.get_entity_folder(id)

    def download_entity_files(self, id = None):
        return self.lambda_framework.download_entity_files(id)

    def download_entity_file(self, filename, id = None):
        return self.lambda_framework.download_entity_file(filename, id)

    def attach_file(self, filename, id = None):
        self.lambda_framework.attach_file(filename, id)

    def attach_folder(self, folder, id = None):
        self.lambda_framework.attach_folder(folder, id)

    def upload(self, filename, name, id = None):
        self.lambda_framework.Upload(filename, name, id)

    def set_exit_code(self, code):
        self.lambda_framework.set_exit_code(code)

    def get_entity(self, id = None):
        return self.lambda_framework.get_entity(id)

    def get_entity_file_list(self, id = None):
        return self.lambda_framework.get_entity_file_list(id)

    def get_callback_update_type(self):
        return self.lambda_framework.get_callback_update_type()

    def get_callback_entity_id(self):
        return self.lambda_framework.get_callback_entity_id()

    def get_callback_entity_state(self):
        return self.lambda_framework.get_callback_entity_state()
    
    def increment_counter(self, counter, offset):
        return self.lambda_framework.increment_counter(counter, offset)
    
    def decrement_counter(self, counter, offset):
        return self.lambda_framework.decrement_counter(counter, offset)

    def set_counter(self, counter, value):
        self.lambda_framework.set_counter(counter, value)

    def get_working_dir(self):
        return self.lambda_framework.get_working_dir()
    
    def get_notify_emails(self):
        return self.lambda_framework.get_notify_emails()

    def get_notify_users_url(self):
        return self.lambda_framework.get_notify_users_url()

    def get_workflow_api(self):
        return self.lambda_framework.get_workflow_api()

    def find_product_definition(self, product, workflow):
        if 'id' in workflow and workflow['id'] == product:
            return workflow
        else:
            if 'tracks' in workflow:
                for track in workflow['tracks']:
                    product_definition = self.find_product_definition(product, track)
                    if product_definition:
                        return product_definition

            return None

    def set_workflow_definition(self, workflow_definition):
        self.lambda_framework.log('Create workflow request')
        workflow_request = request(self.lambda_framework)

        self.lambda_framework.log('Find product definition')
        product_definition = self.find_product_definition(workflow_request.product_id, workflow_definition)
        if product_definition:
            workflow_complete = False
            if workflow_request.update_type:
                if 'on_stage_done' in product_definition:
                    self.lambda_framework.log('Execute on_stage_done event')
                    result = product_definition['on_stage_done'](self.vf_api, workflow_request, self)

                    if result and 'success' in result:
                        if result['success']:
                            self.lambda_framework.log('Done for on_stage_done event')
                            if "complete" in result and result["complete"]:
                                workflow_complete = True
                        else:    
                            self.lambda_framework.log('Error in on_stage_done event: ' + result['error_info'])
                    else:
                        workflow_complete = True
                        self.lambda_framework.log('Event on_stage_done event not found')
            else:    
                if 'on_receive_data' in product_definition:
                    self.lambda_framework.log('Execute on_receive_data event')
                    result = product_definition['on_receive_data'](self.vf_api, workflow_request, self)

                    if result and 'success' in result:
                        if result['success']:
                            self.lambda_framework.log('Done for on_receive_data event')

                            if "complete" in result and result["complete"]:
                                workflow_complete = True
                                result = product_definition['on_stage_done'](self.vf_api, workflow_request, self)
                                if result['success']:
                                    self.lambda_framework.log('Done for on_stage_done event')
                                else:    
                                    self.lambda_framework.log('Error in on_stage_done event: ' + result['error_info'])
                        else:    
                            self.lambda_framework.log('Error in on_receive_data event: ' + result['error_info'])
                            return {'success': False, 'error_info': result['error_info']}
                    else:
                        workflow_complete = True
                        self.lambda_framework.log('Event on_stage_done event not found')

            # Auto-activate version
            if workflow_complete and not 'activated' in workflow_request.properties:
                self.lambda_framework.log(f'Updating product version_active folder...{workflow_request.product_folder_id}')
                product_entity = self.lambda_framework.get_entity(workflow_request.product_folder_id)

                if product_entity and product_entity.ContainsKey('version_active'):
                    previous_active_version = product_entity['version_active'] 
                    if previous_active_version == '':
                        previous_active_version = '0'
                    self.lambda_framework.log(f'Previously active version: {previous_active_version}')
                    result = self.vf_api.update_entity(
                        project=workflow_request.project_id,
                        id=workflow_request.product_folder_id,
                        fields={
                            'version_active' : str(workflow_request.version_folder_id)
                        }
                    )

                    if not result.success:
                        return {'success': False, 'error_info': result.error_info}
                    workflow_request.properties['activated'] = '1'
                    self.lambda_framework.log('Updated product active_version.')        
                else:
                    self.lambda_framework.log(f'product_entity: {workflow_request.product_folder_id} not found')

            # write properties
            self.lambda_framework.log('Saving properties...')
            updatedProperties = {}
            for prop in workflow_request.properties:
                version_property_id = 'version_' + prop
                updatedProperties[version_property_id] = workflow_request.properties[prop]
            self.vf_api.update_entity(
                project=workflow_request.project_id,
                id=workflow_request.version_folder_id,
                fields=updatedProperties  
            )                          
        else:
            self.lambda_framework.log('Product definition not found')

        self.lambda_framework.log('Done')
        return {'success': True}

    def get_product_singleton(self, product_id, singleton_id):
        self.lambda_framework.log(f'getting {singleton_id} for {product_id}...')

        entity = self.lambda_framework.get_entity()
        if entity and entity.ContainsKey('project'):
            project_id = entity['project']
            project_entity = self.lambda_framework.get_entity(project_id, project_id)

            field_id = f'workflow_folder_{product_id}'
            if project_entity and project_entity.ContainsKey(field_id):
                product_folder_id = project_entity[field_id]
                self.lambda_framework.log(f'product_folder_id:{product_folder_id}')

                if product_entity:
                    product_entity = self.vf_api.get_entity(product_folder_id, self.project_id)
                    self.lambda_framework.log(f'product_entity: {product_entity}')

                    singleton_property = f'workflow_singleton_{singleton_id}'
                    if product_entity and product_entity.ContainsKey(singleton_property):
                        return product_entity[singleton_property]
        else:
            self.lambda_framework.log('Project not found in entity.')

        return None

    def get_product_property(self, product_id, property):
        self.lambda_framework.log(f'getting {product_id} property:{property}...')

        entity = self.lambda_framework.get_entity()
        if entity and entity.ContainsKey('project'):
            project_id = entity['project']
            project_entity = self.lambda_framework.get_entity(project_id, project_id)

            if project_entity and project_entity.ContainsKey('workflow_folder_' + product_id):
                product_folder_id = project_entity['workflow_folder_' + product_id]
                product_entity = self.lambda_framework.get_entity(product_folder_id, project_id)
                if product_entity and product_entity.ContainsKey('version_active'):
                    active_version_id = product_entity['version_active']
                    active_version = self.lambda_framework.get_entity(active_version_id, project_id)
                    extended_prop = 'version_' + property
                    if active_version and active_version.ContainsKey(extended_prop):
                        return active_version[extended_prop]
                    else:
                        self.lambda_framework.log('Problem with entity active_version ' + active_version_id)
                else:
                    self.lambda_framework.log('Problem with entity product_entity ' + product_folder_id)
            else:
                self.lambda_framework.log('Problem with entity project ' + project_id)
        else:
            self.lambda_framework.log('Project not found in entity.')

        return None

    def get_parameter_dataframe(self, product_id):
        attribute_product = self.get_product_property(product_id, 'report_entity')
        if attribute_product:
            attribute_file = self.lambda_framework.download_entity_file('report.csv', attribute_product)
            if os.path.isfile(attribute_file):
                return pd.read_csv(attribute_file)
            else:
                self.lambda_framework.log('Parameter file not found.')
        else:
            self.lambda_framework.log('Attribute product not found.')

        return pd.DataFrame()

    def publish_feed(self, entity_id, timestamp, timestamp_id, spatialId, objectId, defaultObjectId, propertyIds, constant_spatial_id, constant_object_id):
        self.lambda_framework.log(f'Publish feed calling timeseries')
        result = self.lambda_framework.publish_feed(entity_id, timestamp, timestamp_id, spatialId, objectId, defaultObjectId, propertyIds, constant_spatial_id, constant_object_id)
        if result == '':
            return {'success': True}
        else:
            return {'success': False, 'error_info': result}

    def publish_reports(self, workflow_request : request, report_id:str, spatialId: str, objectId: str, propertyIds:tuple):
        self.lambda_framework.log(f'Publish reports calling timeseries')
        mesh_id = workflow_request.properties['mesh_id'] 
        self.lambda_framework.log(f'mesh_id:{mesh_id}')
        self.lambda_framework.log(f'report_id:{report_id}')

        self.lambda_framework.log('properties')
        self.lambda_framework.log(propertyIds)

        self.lambda_framework.log(f'Calling process entity to publish report {report_id}...')
        if mesh_id != "":
            mesh_entity = self.vf_api.get_entity(mesh_id, workflow_request.project_id)
            file_date = mesh_entity['file_date']
            self.lambda_framework.log(f'mesh_entity.file_date:{file_date}')
            iso_time = datetime.datetime.fromtimestamp(int(file_date)/1000.0).isoformat()
            self.lambda_framework.log(f'mesh_entity.iso_time:{iso_time}')

            apiResult = self.publish_feed(
                entity_id=report_id, 
                timestamp=iso_time, 
                timestamp_id='', 
                spatialId='', 
                objectId='', 
                defaultObjectId='', 
                propertyIds=propertyIds, 
                constant_spatial_id=spatialId, 
                constant_object_id=objectId)    

            success = apiResult['success']
            if success:
                self.lambda_framework.log('Done publishing feed')    
                return {'success': True}
            else:
                return {'success': False, 'error_info': apiResult.error_info}
        else:
            self.lambda_framework.log(f'mesh_id was empty.')    
            return {'success': False, 'error_info': 'mesh_id was empty.'}

    def get_file_path(self, file):
        if os.path.exists(file):
            return file
        else:
            script_dir = self.lambda_framework.get_working_dir()
            self.lambda_framework.log('file is a relative path')
            file_path = os.path.join(script_dir, file)

            if os.path.exists(file_path):
                return file_path
    
        return None

    def load_file(self, file):
        file_path = self.get_file_path(file)

        if os.path.exists(file_path):
            return open(file_path)
    
        return None

    def create_view(self, workflow_request : request, name : str, view_type : str, view_lambda : str, inputs : dict, props:dict):
        self.lambda_framework.log(f'create_view:name:{name}|view_type:{view_type}|view_lambda:{view_lambda}|inputs:{inputs}|props:{props}')

        if view_type == None:
            lambda_file = self.load_file(view_lambda)
            if lambda_file == None:
                return {'success': False, 'error_info': 'Lambda file not found'}

            result = self.vf_api.create_lambda_python(
                project=workflow_request.project_id, 
                type=self.vf_api.lambda_type.View,
                name=name, 
                fields={
                    'file_folder': workflow_request.version_folder_id,
                    'virtual': '1'
                },
                code=lambda_file.read())
            if not result.success:
                return {'success': False, 'error_info': result.error_info}
            view_type = result.id

        input_fields = {
                'file_folder' : workflow_request.version_folder_id,
                'view_type' : view_type,
                'virtual' : '1',
                'state' : 'COMPLETE',
                'color_legend_attribute' : '',
                'color_legend_attribute_index' : '-1',
                'color_legend_gradient' : 'isoluminant_cgo_70_c39_n256',
                'color_legend_interpolate_gradient' : '1',
                'color_legend_mode' : '2',
                'color_legend_range_max' : '100',
                'color_legend_range_min' : '0',
                'color_legend_range_step' : '1',
                'color_legend_reverse_gradient' : '0',
                'file_date' : str(1000 * int(time.time())),
                'file_type' : 'VIEW',
                'input_filter_colors' : '0',
                'input_filter_e' : '8',
                'input_filter_normals' : '0',
                'input_label_colors' : 'Use Ortho-imagery',
                'input_label_e' : 'Terrain',
                'input_label_normals' : 'Use high resolution detail',
                'input_type_colorlegend' : '7',
                'input_type_colors' : '6',
                'input_type_e' : '3',
                'input_type_normals' : '6',
                'input_value_colors' : '0',
                'input_value_normals' : '0',         
            }
            
        for key in inputs:
            input_fields['input_value_' + key] = inputs[key]

        for key in props:
            input_fields[key] = props[key]
            
        result = self.vf_api.create_entity_raw(
            project=workflow_request.project_id,
            type=self.vf_api.entity_type.View,
            name=name,
            fields=input_fields,
            crs={}
        )

        if not result.success:
            return {'success': False, 'error_info': result.error_info}
        view_object = result.id
        
        self.lambda_framework.log(f'created_view:name:{name}|view_object:{view_object}')

        result = self.vf_api.create_entity_raw(
            project=workflow_request.project_id,
            type=self.vf_api.entity_type.View,
            name=name,
            fields={
                'file_folder' : workflow_request.version_folder_id,
                'view_type' : 'container',
                'state' : 'COMPLETE',
                'entity_container' : view_object
            },
            crs={}
        )

        if not result.success:
            return {'success': False, 'error_info': result.error_info}
        return {'success': True, 'id': result.id, 'error_info': 'None'}

    def create_report(self, workflow_request : request, name : str, report_lambda : str, region : str, lod : int, inputs : dict, fields : dict = None, update_type : str = None):
        lambda_file = self.load_file(report_lambda)
        if lambda_file == None:
            return {'success': False, 'error_info': 'Lambda file not found'}
        
        result = self.vf_api.create_lambda_python(
            project=workflow_request.project_id, 
            type=self.vf_api.lambda_type.Report,
            name=f"Lambda for: {name}", 
            fields={
                'virtual': '1',
                'file_folder': workflow_request.version_folder_id
            },
            code=lambda_file.read())
        
        if not result.success:
            return {'success': False, 'error_info': result.error_info}
        
        report_lambda_id = result.id

        if fields == None:
            fields = {}

        fields['file_folder'] = workflow_request.version_folder_id

        if update_type:
            fields['workflow_update_type'] = workflow_request.get_callback(update_type)

        result = self.vf_api.create_report(
            project=workflow_request.project_id, 
            program=report_lambda_id, 
            region=region,
            lod=str(lod),
            name=name, 
            fields=fields,
            inputs=inputs)
        
        if not result.success:
            return {'success': False, 'lambda_id': report_lambda_id, 'error_info': result.error_info}
        
        return {'success': True, 'id': result.id, 'lambda_id': report_lambda_id, 'error_info': 'None'}

    def create_lambda(self, workflow_request : request, name : str, type : str, lambda_code : str):
        lambda_file = self.load_file(lambda_code)
        if lambda_file == None:
            return {'success': False, 'error_info': 'Lambda file not found'}

        result = self.vf_api.create_lambda_python(
            project=workflow_request.project_id, 
            type=type,
            name=name, 
            fields={
                'file_folder': workflow_request.version_folder_id
            },
            code=lambda_file.read())
        
        if not result.success:
            return {'success': False, 'error_info': result.error_info}
        
        report_lambda_id = result.id
        return {'success': True, 'id': report_lambda_id}

    def create_export(self, workflow_request : request, name : str, export_lambda : str, region : str, lod : int, inputs : dict, update_type : str = None):
        lambda_file = self.load_file(export_lambda)
        if lambda_file == None:
            return {'success': False, 'error_info': 'Lambda file not found'}

        result = self.vf_api.create_lambda_python(
            project=workflow_request.project_id, 
            type=self.vf_api.lambda_type.Report,
            name="Export Lambda for " + name, 
            fields={
                'virtual': '1',
                'file_folder': workflow_request.version_folder_id
            },
            code=lambda_file.read())
        if not result.success:
            return {'success': False, 'error_info': result.error_info}
        report_lambda_id = result.id

        if fields == None:
            fields = {}

        fields['file_folder'] = workflow_request.version_folder_id
        fields['export_type'] = 'mesh'

        if update_type:
            fields['workflow_update_type'] = workflow_request.get_callback(update_type)

        result = self.vf_api.create_export(
            project=workflow_request.project_id, 
            program=report_lambda_id, 
            region=region,
            lod=str(lod),
            name=name, 
            fields=fields,
            inputs=inputs)

        if not result.success:
            return {'success': False, 'lambda_id': report_lambda_id, 'error_info': result.error_info}

        return {'success': True, 'id': result.id, 'lambda_id': report_lambda_id, 'error_info': 'None'}

    def get_meta_attributes(self, workflow_request : request, projectId:str):
        attribute_product = workflow_request.get_product_property('META_ATTR_BM', 'report_entity')
        self.lambda_framework.log(f'attribute_product:{attribute_product}')
        file_name = 'report.csv'
        local_attr_file = os.path.join(workflow_request.scrap_folder, file_name)
        self.lambda_framework.log(f'Downloading {file_name} to {local_attr_file}...')
        if not self.vf_api.download_file(projectId, attribute_product, file_name, local_attr_file):
            return {'success': False, 'error_info': f'Could not download {file_name} to {local_attr_file}'}
        self.lambda_framework.log(f'Downloaded {file_name} to {local_attr_file}...')
        
        self.lambda_framework.log(f'Processing {local_attr_file}...')
        
        csvfile = open(f"{local_attr_file}", mode='r', encoding='utf-8-sig' )
        reader = csv.DictReader(csvfile)
        record_count = 0
        view_attribute = ''
        attributes_arr = []
        groups_arr = []
        grades_arr = []
        for row in reader:
            #self.lambda_framework.log(f'row:{row}')
            ID = row['Property']
            Report = row['Report']
            View = row['View']
            Group = row['Group']
            Grade = row['Grade']
            if View == '1' and view_attribute == '':
                self.lambda_framework.log(f'Assigning attribute: {ID} with index: {record_count}')
                view_attribute = ID
            if Report == '1':
                attributes_arr.append(ID)
            if Group == '1':
                groups_arr.append(ID)
            if Grade == '1':
                grades_arr.append(ID)
            record_count += 1

        attributes = ",".join(attributes_arr)
        groups = ",".join(groups_arr)
        grades = ",".join(grades_arr)

        self.lambda_framework.log(f'attributes:{attributes}')
        self.lambda_framework.log(f'groups:{groups}')
        self.lambda_framework.log(f'grades:{grades}') 

        return {'attributes':attributes, 'groups': groups, 'grades':grades, 'view_attribute': view_attribute}

    def is_number(element) -> bool:
        try:
            float(element)
            return True
        except:
            return False
        
    def notify_users(self, file_entity_id):
        self.lambda_framework.log(f'notify_users for file_entity_id: {file_entity_id}')
        file_entity = self.lambda_framework.get_entity(file_entity_id)
        if file_entity != None:
            emails_to = self.lambda_framework.get_notify_emails()
            if emails_to is None:
                self.lambda_framework.log('No email were sent because OS getenv variable vf_notify_emails is None')
                return { "success": True }
            
            self.lambda_framework.log(f'emails_to: {emails_to}')

            emails_to = emails_to.split(",")
            folder_entity = self.lambda_framework.get_entity(file_entity['file_folder'])
            parent_entity = self.lambda_framework.get_entity(folder_entity['file_folder'])
            
            if 'user' in file_entity and file_entity['user'] not in emails_to:
                emails_to.append(file_entity['user'])
            if 'user' in folder_entity and folder_entity['user'] not in emails_to:
                emails_to.append(folder_entity['user'])
            if 'user' in parent_entity and parent_entity['user'] not in emails_to:
                emails_to.append(parent_entity['user'])

            self.lambda_framework.log(f'notify users: {emails_to}')

            url = self.lambda_framework.get_notify_users_url()
            self.lambda_framework.log(f'notify url: {url}')

            if file_entity["state"] == "COMPLETE":
                subject = f"Completed {file_entity['name']} in {parent_entity['name']} ({folder_entity['name']})"
                body = f"Dear team member,\n\nThe following item has been completed:\n\n    {file_entity['name']} in {parent_entity['name']} ({folder_entity['name']})\n\nYou can access this item here:\n\n{self.lambda_framework.get_workflow_api()}app/catalog/{file_entity_id}\n\nPlease do not reply to this message. It has been generated by an automated workflow."
            else:
                subject = f"Error in {file_entity['name']} in {parent_entity['name']} ({folder_entity['name']})"
                body = f"Dear team member,\n\nWe have encountered some issues while processing the following item:    \n\n{file_entity['name']} in {parent_entity['name']} ({folder_entity['name']})\n\n For more information, please access the log tab of the item. You can access this item here:\n\n{self.lambda_framework.get_workflow_api()}app/catalog/{file_entity_id}\n\nPlease do not reply to this message. It has been generated by an automated workflow."

            self.lambda_framework.log(f'subject: {subject}')
            self.lambda_framework.log(f'body: {body}')

            data = {
                "from": "noreply@voxelfarm.com",
                "to": ",".join(emails_to),
                "subject": subject,
                "body": body
            }
            headers = {"Content-type": "application/json", "Accept": "text/plain"}

            req = requests.post(url, data=json.dumps(data), headers=headers)
            if req.status_code >= 200 and req.status_code < 300:
                self.lambda_framework.log(f'Emails sent successfully')
                return { "success": True }
            else:
                self.lambda_framework.log(f'Error sending the emails')
                return { "success": False }
        else:
            self.lambda_framework.log(f'Entity : {file_entity_id} not found')
            return { "success": False }

    def process_entity(self, workflow_request, type, name, fields = None, update_type = None):

        if fields == None:
            fields = {}

        fields['file_folder'] = workflow_request.version_folder_id

        if update_type:
            fields['workflow_update_type'] = workflow_request.get_callback(update_type)

        return self.vf_api.create_entity_processed(
            project=workflow_request.project_id, 
            type=type, 
            name=name, 
            fields=fields, 
            crs=workflow_request.crs, 
            callback=None)

    def process_lambda_entity(self, workflow_request, name, inputs, code, files, fields = None, update_type = None):
        verified_files = []
        if files:
            for file in files:
                verified_files.append(self.get_file_path(file))

        if fields == None:
            fields = {}

        fields['file_folder'] = workflow_request.version_folder_id

        if update_type:
            fields['workflow_update_type'] = workflow_request.get_callback(update_type)

        return self.vf_api.create_python_process_entity(
            project=workflow_request.project_id, 
            name=name, 
            fields=fields, 
            inputs=inputs,
            code=code,
            files=verified_files, 
            callback=None)










import voxelfarm

class process_lambda_framework:

    def __init__(self):  
        pass

    def input_string(self, id, label, default = ""):
        return ""

    def log(self, message):
        pass

    def progress(self, progress, message):
        pass

    def get_cfg(self, name):
        return ""

    def get_scrap_folder(self):
        return ""

    def get_tools_folder(self):
        return ""

    def get_entity_folder(self, id = None):
        return ""

    def download_entity_files(self, id = None):
        return ""

    def download_entity_file(self, filename, id = None):
        return ""

    def attach_file(self, filename, id = None):
        pass

    def attach_folder(self, folder, id = None):
        pass

    def upload(self, filename, name, id = None):
        pass

    def set_exit_code(self, code):
        pass

    def get_entity(self, id = None):
        return None

    def get_entity_file_list(self, id = None):
        return []

    def export_file(self, local_file_location, drop_zone_file_location):
        pass

class process_lambda_host:

    def __init__(self, framework = None):  
        if framework:
            self.lambda_framework = framework
        else:
            if voxelfarm.voxelfarm_framework:
                self.lambda_framework = voxelfarm.voxelfarm_framework
            else:
                self.lambda_framework = process_lambda_framework()

    def input_string(self, id, label, default = ""):
        return self.lambda_framework.input_string(id, label, default)

    def log(self, message):
        self.lambda_framework.log(message)

    def progress(self, progress, message):
        self.lambda_framework.progress(progress, message)

    def get_cfg(self, name):
        return self.lambda_framework.GetCfg(name)

    def get_scrap_folder(self):
        return self.lambda_framework.get_scrap_folder()

    def get_tools_folder(self):
        return self.lambda_framework.get_tools_folder()

    def get_entity_folder(self, id = None):
        return self.lambda_framework.get_entity_folder(id)

    def download_entity_files(self, id = None):
        return self.lambda_framework.download_entity_files(id)

    def download_entity_file(self, filename, id = None):
        return self.lambda_framework.download_entity_file(filename, id)

    def attach_file(self, filename, id = None):
        self.lambda_framework.attach_file(filename, id)

    def attach_folder(self, folder, id = None):
        self.lambda_framework.attach_folder(folder, id)

    def upload(self, filename, name, id = None):
        self.lambda_framework.Upload(filename, name, id)

    def set_exit_code(self, code):
        self.lambda_framework.set_exit_code(code)

    def get_entity(self, id = None):
        return self.lambda_framework.get_entity(id)

    def get_entity_file_list(self, id = None):
        return self.lambda_framework.get_entity_file_list(id)

    def export_file(self, local_file_location, drop_zone_file_location):
        return self.lambda_framework.export_file(local_file_location, drop_zone_file_location)

